import math

import ee

from . import calcs


def lazy_property(fn):
    """Decorator that makes a property lazy-evaluated

    https://stevenloria.com/lazy-properties/
    """
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


class Daily():
    def __init__(self, tmax, tmin, ea, rs, uz, zw, elev, lat, doy, method='asce', rso_type=None, rso=None):
        """ASCE Daily Standardized Reference Evapotranspiration (ET)

        Arguments
        ---------
        tmax : ee.Image
            Maximum daily temperature [C].
        tmin : ee.Image
            Minimum daily temperature [C].
        ea : ee.Image
            Actual vapor pressure [kPa].
        rs : ee.Image
            Incoming shortwave solar radiation [MJ m-2 day-1].
        uz : ee.Image
            Wind speed [m s-1].
        zw : ee.Number
            Wind speed height [m].
        elev : ee.Image or ee.Number
            Elevation [m].
        lat : ee.Image or ee.Number
            Latitude [degrees].
        doy : ee.Number
            Day of year.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1].
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple', 'array'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation
            * 'array' -- Read Rso values from "rso" function parameter
        rso : ee.Image, ee.Number, or None, optional
            Clear sky solar radiation [MJ m-2 day-1] (the default is None).
            Only used if rso_type == 'array'.

        Raises
        ------
        ValueError
            If 'method' or 'rso_type' parameter is invalid.

        Notes
        -----
        Latitude units are degrees, not radians.

        References
        ----------
        .. [1] ASCE-EWRI (2005). The ASCE standardized reference evapotranspiration
            equation. ASCE-EWRI Standardization of Reference Evapotranspiration
            Task Committee Rep., ASCE Reston, Va.

        """

        if method.lower() not in ['asce', 'refet']:
            raise ValueError('method must be "asce" or "refet"')

        if rso_type is None:
            pass
        elif rso_type.lower() not in ['simple', 'full', 'array']:
            raise ValueError('rso_type must be None, "simple", "full", or "array')
        elif rso_type.lower() in 'array':
            # Check that rso is an ee.Image or ee.Number?
            pass

        # Get time_start from tmin
        # Should time_start be set in init?
        self.time_start = ee.Image(tmin).get('system:time_start')
        self.date = ee.Date(self.time_start)

        # Do these all need to be set onto self?
        self.tmax = tmax
        self.tmin = tmin
        self.ea = ea
        self.rs = rs
        self.uz = uz
        self.zw = zw
        self.elev = elev
        self.lat = lat
        self.doy = doy

        # Convert latitude to radians
        self.lat = self.lat.multiply(math.pi / 180)

        # To match standardized form, pair is calculated from elevation
        self.pair = calcs.air_pressure(self.elev, method)

        # Psychrometric constant (Eq. 4)
        self.psy = self.pair.multiply(0.000665)

        self.tmean = self.tmax.add(self.tmin).multiply(0.5)
        self.es_slope = calcs.es_slope(self.tmean, method)

        # Saturated vapor pressure
        self.es = calcs.sat_vapor_pressure(self.tmax).add(calcs.sat_vapor_pressure(self.tmin)).multiply(0.5)

        # Vapor pressure deficit
        self.vpd = calcs.vpd(es=self.es, ea=self.ea)

        # Extraterrestrial radiation
        self.ra = calcs.ra_daily(lat=self.lat, doy=self.doy, method=method)

        # Clear sky solar radiation
        # If rso_type is not set, use the method
        # If rso_type is set, use rso_type directly
        if rso_type is None:
            if method.lower() == 'asce':
                self.rso = calcs.rso_simple(ra=self.ra, elev=self.elev)
            elif method.lower() == 'refet':
                self.rso = calcs.rso_daily(ea=self.ea, pair=self.pair, ra=self.ra, doy=self.doy, lat=self.lat)
        elif rso_type.lower() == 'simple':
            self.rso = calcs.rso_simple(ra=self.ra, elev=self.elev)
        elif rso_type.lower() == 'full':
            self.rso = calcs.rso_daily(ea=self.ea, pair=self.pair, ra=self.ra, doy=self.doy, lat=self.lat)
        elif rso_type.lower() == 'array':
            # Use rso array passed to function
            self.rso = rso

        # Cloudiness fraction
        self.fcd = calcs.fcd_daily(rs=self.rs, rso=self.rso)

        # Net long-wave radiation
        self.rnl = calcs.rnl_daily(tmax=self.tmax, tmin=self.tmin, ea=self.ea, fcd=self.fcd)

        # Net radiation
        self.rn = calcs.rn(self.rs, self.rnl)

        # Wind speed
        self.u2 = calcs.wind_height_adjust(uz=self.uz, zw=self.zw)

    def etsz(self, surface):
        """Standardized reference ET

        Parameters
        ----------
        surface : {'alfalfa', 'etr', 'tall', 'grass', 'eto', 'short'}
            Reference surface type.

        Returns
        -------
        ee.Image

        """
        if surface.lower() in ['alfalfa', 'etr', 'tall']:
            return self.etr
        elif surface.lower() in ['grass', 'eto', 'short']:
            return self.eto
        else:
            raise ValueError('unsupported surface type: {}'.format(surface))

    @lazy_property
    def eto(self):
        """Short (grass) reference surface"""
        self.cn = 900
        self.cd = 0.34
        return ee.Image(self._etsz().rename(['eto']).set('system:time_start', self.time_start))

    @lazy_property
    def etr(self):
        """Tall (alfalfa) reference surface"""
        self.cn = 1600
        self.cd = 0.38
        return ee.Image(self._etsz().rename(['etr']).set('system:time_start', self.time_start))

    def _etsz(self):
        """Daily reference ET (Eq. 1)

        Returns
        -------
        etsz : ee.Image
            Standardized reference ET [mm].

        """
        return (
            self.u2.multiply(self.vpd).multiply(self.cn).multiply(self.psy).divide(self.tmean.add(273))
            .add(self.es_slope.multiply(self.rn).multiply(0.408))
            .divide(self.u2.multiply(self.cd).add(1).multiply(self.psy).add(self.es_slope))
        )
        # # CGM - Intentionally leaving expression form of calculation here for reference
        # return self.tmin.expression(
        #     '((0.408 * es_slope * rn) + (psy * cn * u2 * vpd / (tmean + 273))) / '
        #     '(es_slope + psy * (cd * u2 + 1))',
        #     {
        #         'cd': self.cd, 'cn': self.cn, 'es_slope': self.es_slope,
        #         'psy': self.psy, 'rn': self.rn, 'tmean': self.tmean,
        #         'u2': self.u2, 'vpd': self.vpd,
        #     }
        #  )

    @lazy_property
    def etw(self):
        """Priestley-Taylor evaporation (alpha = 1.26)

        Returns
        -------
        etw : ee.Image
            Priestley-Taylor ET [mm].

        References
        ----------
        https://wetlandscapes.github.io/blog/blog/penman-monteith-and-priestley-taylor/

        """
        return (
            self.es_slope
            .expression(
                '(alpha * es_slope * rn * 1000 / (2453 * (es_slope + psy)))',
                {'es_slope': self.es_slope, 'rn': self.rn, 'psy': self.psy, 'alpha': 1.26}
            )
            .rename(['etw'])
            .set('system:time_start', self.time_start)
        )

    @lazy_property
    def eto_fs1(self):
        """UF-IFAS Extension FS1 Radiation Term (ETrad)

        Returns
        -------
        eto_fs1 : ee.Image
            FS1 ETrad [mm].

        References
        ----------
        https://edis.ifas.ufl.edu/pdffiles/ae/ae45900.pdf

        """
        return (
            self.u2
            .expression(
                '(delta / (delta + psy * (1 + 0.34 * u2))) * (0.408 * rn)',
                {'delta': self.es_slope, 'psy': self.psy, 'u2': self.u2, 'rn': self.rn}
            )
            .rename(['eto_fs1'])
            .set('system:time_start', self.time_start)
        )

    @lazy_property
    def eto_fs2(self):
        """UF-IFAS Extension FS2 Wind Term (ETwind)

        Returns
        -------
        eto_fs2 : ee.Image
            FS2 ETwind [mm].

        References
        ----------
        https://edis.ifas.ufl.edu/pdffiles/ae/ae45900.pdf

        """
        # Temperature Term (Eq. 14)
        tt = self.u2.expression('(900 / (t + 273)) * u2', {'t': self.tmean, 'u2': self.u2})

        # Psi Term (Eq. 13)
        pt = self.u2.expression(
            'psy / (slope + psy * (1 + 0.34 * u2))',
            {'slope': self.es_slope, 'psy': self.psy, 'u2': self.u2}
        )

        return (
            self.es.subtract(self.ea).multiply(pt).multiply(tt)
            .rename(['eto_fs2'])
            .set('system:time_start', self.time_start)
        )

    @lazy_property
    def pet_hargreaves(self):
        """Hargreaves potential ET

        Returns
        -------
        hargreaves_pet : ee.Image
            Hargreaves ET [mm].

        References
        ----------

        """
        return (
            self.tmax
            .expression(
                '0.0023 * (tmean + 17.8) * ((tmax - tmin) ** 0.5) * 0.408 * ra',
                {'tmean': self.tmean, 'tmax': self.tmax, 'tmin': self.tmin, 'ra': self.ra}
            )
            .rename(['pet_hargreaves'])
            .set('system:time_start', self.time_start)
        )

    @classmethod
    def gridmet(cls, input_img, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from a GRIDMET image

        Parameters
        ----------
        input_img : ee.Image
            GRIDMET image from the collection IDAHO_EPSCOR/GRIDMET.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The standard GRIDMET elevation image
            (projects/openet/assets/meteorology/gridmet/elevation) will be used if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The latitude will be computed
            dynamically using ee.Image.pixelLonLat() if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Temperatures are converted from K to C.
        Solar radiation is converted from W m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from specific humidity (GRIDMET sph)
            and air pressure (from elevation).

        """
        image_date = ee.Date(input_img.get('system:time_start'))

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            elev = ee.Image('projects/openet/assets/meteorology/gridmet/elevation')
        if lat is None:
            lat = ee.Image('projects/openet/assets/meteorology/gridmet/latitude')

        return cls(
            tmax=input_img.select(['tmmx']).subtract(273.15),
            tmin=input_img.select(['tmmn']).subtract(273.15),
            ea=calcs.actual_vapor_pressure(
                pair=calcs.air_pressure(elev, method),
                q=input_img.select(['sph']),
            ),
            rs=input_img.select(['srad']).multiply(0.0864),
            uz=input_img.select(['vs']),
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(image_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def maca(cls, input_img, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from a MACA image

        Parameters
        ----------
        input_img : ee.Image
            MACA image from the collection IDAHO_EPSCOR/MACAv2_METDATA.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The standard GRIDMET elevation image
            (projects/openet/assets/meteorology/gridmet/elevation) will be used if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The latitude will be computed
            dynamically using ee.Image.pixelLonLat() if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Temperatures are converted from K to C.
        Solar radiation is converted from W m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from specific humidity (MACA huss)
            and air pressure (from elevation).
        Wind speed from east and north vector components m s-1.

        """
        image_date = ee.Date(input_img.get('system:time_start'))

        # TODO: CHECK MACA transform (Why the difference?)
        # maca_transform = [0.04163593073593073, 0, -124.7722,
        #                   0, 0.04159470085470086, 25.0631]
        # gridmet_transform = [0.041666666666666664, 0, -124.78749996666667,
        #                      0, -0.041666666666666664, 49.42083333333334]

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            # TODO: should we use gridmet elev? maca grid is shifted
            elev = ee.Image('projects/openet/assets/meteorology/gridmet/elevation')
        if lat is None:
            elev = ee.Image('projects/openet/assets/meteorology/gridmet/latitude')

        def wind_magnitude(input_img):
            """Compute daily wind magnitude from vectors"""
            return (
                ee.Image(input_img.select(['uas'])).pow(2)
                .add(ee.Image(input_img.select(['vas'])).pow(2))
                .sqrt().rename(['uz'])
            )

        return cls(
            tmax=input_img.select(['tasmax']).subtract(273.15),
            tmin=input_img.select(['tasmin']).subtract(273.15),
            ea=calcs.actual_vapor_pressure(
                pair=calcs.air_pressure(elev, method),
                q=input_img.select(['huss']),
            ),
            rs=input_img.select(['rsds']).multiply(0.0864),
            uz=ee.Image(wind_magnitude(input_img)).select(['uz']),
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(image_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def nldas(cls, input_coll, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from a hourly NLDAS image collection

        Parameters
        ----------
        input_coll : ee.ImageCollection
            Collection of NLDAS hourly images for a single day from the
            collection NASA/NLDAS/FORA0125_H002.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  A custom NLDAS elevation image
            (projects/eddi-noaa/nldas/elevation) will be used if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The latitude will be computed
            dynamically using ee.Image.pixelLonLat() if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Solar radiation is converted from W m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from specific humidity and air
            pressure (from elevation).

        """
        input_coll = ee.ImageCollection(input_coll)
        image_date = ee.Date(ee.Image(input_coll.first()).get('system:time_start'))

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            elev = ee.Image('projects/openet/assets/meteorology/nldas/ancillary/elevation')
        if lat is None:
            lat = ee.Image('projects/openet/assets/meteorology/nldas/ancillary/latitude')

        def wind_magnitude(input_img):
            """Compute hourly wind magnitude from vectors"""
            return (
                ee.Image(input_img.select(['wind_u'])).pow(2)
                .add(ee.Image(input_img.select(['wind_v'])).pow(2))
                .sqrt().rename(['uz'])
            )

        return cls(
            tmax=input_coll.select(['temperature']).max(),
            tmin=input_coll.select(['temperature']).min(),
            ea=calcs.actual_vapor_pressure(
                pair=calcs.air_pressure(elev, method),
                q=input_coll.select(['specific_humidity']).mean(),
            ),
            rs=input_coll.select(['shortwave_radiation']).sum().multiply(0.0036),
            uz=ee.Image(ee.ImageCollection(input_coll.map(wind_magnitude)).mean()),
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(image_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def cfsv2(cls, input_coll, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from a 6-hourly CFSv2 image collection

        Parameters
        ----------
        input_coll : ee.ImageCollection
            Collection of CFSv2 6 hourly images for a single day from the
            collection NOAA/CFSV2/FOR6H.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The SRTM elevation image (CGIAR/SRTM90_V4)
            will be reprojected to the CFSv2 grid if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The latitude will be computed
            dynamically using ee.Image.pixelLonLat() if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Temperatures are converted from K to C.
        Solar radiation is converted from W m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from specific humidity and air
            pressure (from elevation).

        """
        input_coll = ee.ImageCollection(input_coll)
        image_date = ee.Date(ee.Image(input_coll.first()).get('system:time_start'))

        cfsv2_crs = 'EPSG:4326'
        # Transform for 2011 to Present
        cfsv2_transform = [0.20454520376789903, 0., -180, 0., -0.20442210122586923, 90]
        # Transform for 1979 to 2010
        # cfsv2_transform = [0.3124995757764987, 0, -180, 0, -0.31221217943274454, 90]

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            # TODO: Build a CFSv2 elevation asset
            elev = ee.Image('CGIAR/SRTM90_V4').reproject(cfsv2_crs, cfsv2_transform)
        if lat is None:
            lat = ee.Image.pixelLonLat().select('latitude').reproject(cfsv2_crs, cfsv2_transform)

        def wind_magnitude(input_img):
            """Compute hourly wind magnitude from vectors"""
            u_img = ee.Image(input_img).select(['u-component_of_wind_height_above_ground'])
            v_img = ee.Image(input_img).select(['v-component_of_wind_height_above_ground'])
            return u_img.pow(2).add(v_img.pow(2)).sqrt()

        return cls(
            tmax=input_coll.select(['Maximum_temperature_height_above_ground_6_Hour_Interval'])
                .max().subtract(273.15),
            tmin=input_coll.select(['Minimum_temperature_height_above_ground_6_Hour_Interval'])
                .min().subtract(273.15),
            ea=calcs.actual_vapor_pressure(
                pair=calcs.air_pressure(elev, method),
                q=input_coll.select(['Specific_humidity_height_above_ground']).mean()
            ),
            # TODO: Check the conversion on solar
            rs=input_coll.select(['Downward_Short-Wave_Radiation_Flux_surface_6_Hour_Average'])
                .mean().multiply(0.0864),
            uz=ee.Image(ee.ImageCollection(input_coll.map(wind_magnitude)).mean()),
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(image_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def rtma(cls, input_coll, rs=None, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from a hourly RTMA image collection

        Parameters
        ----------
        input_coll : ee.ImageCollection
            Collection of RTMA hourly images for a single day from the
            collection NOAA/NWS/RTMA.
        rs : ee.Image, str, optional
            Incoming solar radiation [MJ m-2 day-1].  The GRIDMET image for the
            concurrent day will be used if not set.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The RTMA elevation image asset will be used if not set
            (projects/openet/assets/meteorology/rtma/ancillary/elevation).
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The RTMA latitude image asset will be used if not set
            (projects/openet/assets/meteorology/rtma/ancillary/latitude).
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Solar radiation is converted from W m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from specific humidity and air
            pressure (from elevation).

        """
        input_coll = ee.ImageCollection(input_coll)
        start_date = ee.Date(ee.Image(input_coll.first()).get('system:time_start'))

        # Parse the solar radiation input
        if isinstance(rs, ee.Image):
            pass
        elif isinstance(rs, ee.Number) or isinstance(rs, float) or isinstance(rs, int):
            rs = ee.Image.constant(rs)
        elif (rs is None) or (rs.upper() == 'GRIDMET'):
            rs_coll = (
                ee.ImageCollection('IDAHO_EPSCOR/GRIDMET')
                .filterDate(start_date, start_date.advance(1, 'day'))
                .select(['srad'])
            )
            rs = ee.Image(rs_coll.first()).multiply(0.0864)
        elif rs.upper() == 'NLDAS':
            rs_coll = (
                ee.ImageCollection('NASA/NLDAS/FORA0125_H002')
                .filterDate(start_date, start_date.advance(1, 'day'))
                .select(['shortwave_radiation'])
            )
            # TODO: Check this unit conversion
            rs = ee.Image(rs_coll.sum()).multiply(0.0036)
        else:
            raise ValueError('Unsupported Rs input')

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            elev = ee.Image('projects/openet/assets/meteorology/rtma/ancillary/elevation')
        if lat is None:
            lat = ee.Image('projects/openet/assets/meteorology/rtma/ancillary/latitude')

        # # Not computing wind speed from the components since wind speed is available
        # def wind_magnitude(input_img):
        #     """Compute hourly wind magnitude from vectors"""
        #     return (
        #         ee.Image(input_img.select(['UGRD'])).pow(2)
        #         .add(ee.Image(input_img.select(['VGRD'])).pow(2))
        #         .sqrt().rename(['WIND'])
        #     )

        return cls(
            tmax=input_coll.select(['TMP']).max(),
            tmin=input_coll.select(['TMP']).min(),
            ea=calcs.actual_vapor_pressure(
                pair=calcs.air_pressure(elev, method),
                q=input_coll.select(['SPFH']).mean()
            ),
            rs=rs,
            # Using wind speed directly instead of computing from components
            uz=input_coll.select(['WIND']).mean(),
            # uz=ee.Image(ee.ImageCollection(input_coll.map(wind_magnitude)).mean())
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(start_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def era5(cls, input_coll, zw=None, elev=None, lat=None, method='asce', rso_type=None):
        """Initialize daily RefET from an hourly ERA5 image collection

        Parameters
        ----------
        input_coll : ee.ImageCollection
            Collection of ERA5 hourly images for a single day from the
            collection ECMWF/ERA5/HOURLY.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The OpenET ERA5 elevation image
            (projects/openet/assets/meteorology/era5/ancillary/elevation)
            will be used if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The OpenET ERA5 latitude image
            (projects/openet/assets/meteorology/era5/ancillary/latitude)
            will be used if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation

        Notes
        -----
        Temperatures are converted from K to C.
        Solar radiation is summed and converted from J m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from dew point temperature.

        """
        input_coll = ee.ImageCollection(input_coll)
        start_date = ee.Date(ee.Image(input_coll.first()).get('system:time_start'))

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            elev = ee.Image('projects/openet/assets/meteorology/era5/ancillary/elevation')
        if lat is None:
            lat = ee.Image('projects/openet/assets/meteorology/era5/ancillary/latitude')

        def wind_magnitude(input_img):
            """Compute hourly wind magnitude from vectors"""
            return (
                ee.Image(input_img.select(['u_component_of_wind_10m'])).pow(2)
                .add(ee.Image(input_img.select(['v_component_of_wind_10m'])).pow(2))
                .sqrt().rename(['wind_10m'])
            )

        return cls(
            tmax=input_coll.select(['temperature_2m']).max().subtract(273.15),
            tmin=input_coll.select(['temperature_2m']).min().subtract(273.15),
            ea=calcs.sat_vapor_pressure(input_coll.select(['dewpoint_temperature_2m']).mean().subtract(273.15)),
            rs=input_coll.select(['surface_solar_radiation_downwards']).sum().divide(1000000),
            uz=ee.Image(ee.ImageCollection(input_coll.map(wind_magnitude)).mean()),
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(start_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    @classmethod
    def era5_land(
            cls,
            input_coll,
            zw=None,
            elev=None,
            lat=None,
            method='asce',
            rso_type=None,
            fill_edge_cells=False,
    ):
        """Initialize daily RefET from an hourly ERA5-Land image collection

        Parameters
        ----------
        input_coll : ee.ImageCollection
            Collection of ERA5-Land hourly images for a single day from the
            collection ECMWF/ERA5_LAND/HOURLY.
        zw : ee.Number, optional
            Wind speed height [m] (the default is 10).
        elev : ee.Image or ee.Number, optional
            Elevation image [m].  The OpenET ERA5-Land elevation image
            (projects/openet/assets/meteorology/era5land/ancillary/elevation)
            will be used if not set.
        lat : ee.Image or ee.Number
            Latitude image [degrees].  The OpenET ERA5-Land latitude image
            (projects/openet/assets/meteorology/era5land/ancillary/latitude)
            will be used if not set.
        method : {'asce' (default), 'refet'}, optional
            Specifies which calculation method to use.
            * 'asce' -- Calculations will follow ASCE-EWRI 2005.
            * 'refet' -- Calculations will follow RefET software.
        rso_type : {None (default), 'full' , 'simple'}, optional
            Specifies which clear sky solar radiation (Rso) model to use.
            * None -- Rso type will be determined from "method" parameter
            * 'full' -- Full clear sky solar formulation
            * 'simple' -- Simplified clear sky solar formulation
        fill_edge_cells : bool, optional
            If True, fill any masked pixels within a one cell buffer of the input image.
            This will fill in small holes and cells along the coasts.

        Notes
        -----
        Temperatures are converted from K to C.
        Solar radiation is summed and converted from J m-2 to MJ m-2 day-1.
        Actual vapor pressure is computed from dew point temperature.

        """
        input_coll = ee.ImageCollection(input_coll)
        start_date = ee.Date(ee.Image(input_coll.first()).get('system:time_start'))

        if zw is None:
            zw = ee.Number(10)
        if elev is None:
            elev = ee.Image('projects/openet/assets/meteorology/era5land/ancillary/elevation')
        if lat is None:
            lat = ee.Image('projects/openet/assets/meteorology/era5land/ancillary/latitude')

        def wind_magnitude(input_img):
            """Compute hourly wind magnitude from vectors"""
            return (
                ee.Image(input_img.select(['u_component_of_wind_10m'])).pow(2)
                .add(ee.Image(input_img.select(['v_component_of_wind_10m'])).pow(2))
                .sqrt().rename(['wind_10m'])
            )

        # Compute the input variables and convert to standard units
        tmax = input_coll.select(['temperature_2m']).max().subtract(273.15)
        tmin = input_coll.select(['temperature_2m']).min().subtract(273.15)
        ea = calcs.sat_vapor_pressure(input_coll.select(['dewpoint_temperature_2m']).mean().subtract(273.15))
        rs = input_coll.select(['surface_solar_radiation_downwards_hourly']).sum().divide(1000000)
        uz = ee.Image(ee.ImageCollection(input_coll.map(wind_magnitude)).mean())

        # Fill any masked pixels along the edge of the land mask
        # This will fill most coastal pixels and any small holes
        if fill_edge_cells:
            def fill(image):
                img = ee.Image(image)
                return img.unmask(
                    img.reduceNeighborhood('mean', ee.Kernel.square(1), 'kernel', False)
                    .reproject(img.projection())
                )
            tmax = fill(tmax)
            tmin = fill(tmin)
            ea = fill(ea)
            rs = fill(rs)
            uz = fill(uz)

        return cls(
            tmax=tmax,
            tmin=tmin,
            ea=ea,
            rs=rs,
            uz=uz,
            zw=zw,
            elev=elev,
            lat=lat,
            doy=ee.Number(start_date.getRelative('day', 'year')).add(1).double(),
            method=method,
            rso_type=rso_type,
        )

    # @classmethod
    # def prism(cls, input_img, lat=None):
    #     """Initialize daily RefET from a PRISM image
    #
    #     Parameters
    #     ----------
    #     input_img : ee.Image
    #         PPRISM image from the collection OREGONSTATE/PRISM/AN81d.
    #     lat : ee.Image or ee.Number
    #         Latitude image [degrees].  The latitude will be computed
    #         dynamically using ee.Image.pixelLonLat() if not set.
    #
    #     Notes
    #     -----
    #     PRISM can only be used to compute Hargreaves PET.
    #
    #     """
    #     image_date = ee.Date(input_img.get('system:time_start'))
    #
    #     if lat is None:
    #         lat = ee.Image('projects/earthengine-legacy/assets/'
    #                        'projects/climate-engine/prism/elevation')\
    #             .multiply(0).add(ee.Image.pixelLonLat().select('latitude'))\
    #             .rename(['latitude'])
    #         # prism_transform = [0.0416666666667, 0, -125.02083333333336,
    #         #                    0, -0.0416666666667, 49.93749999999975]
    #         # lat = ee.Image.pixelLonLat().select('latitude')\
    #         #     .reproject('EPSG:4326', prism_transform)
    #         # lat = input_img.select([0]).multiply(0)\
    #         #     .add(ee.Image.pixelLonLat().select('latitude'))
    #
    #     return cls(
    #         tmax=input_img.select(['tmmx']),
    #         tmin=input_img.select(['tmmn']),
    #         ea=0,
    #         rs=0,
    #         uz=0,
    #         zw=2,
    #         elev=0,
    #         lat=lat,
    #         doy=ee.Number(image_date.getRelative('day', 'year')).add(1).double(),
    #         method='asce',
    #     )



import math

import ee


def air_pressure(elev, method='asce'):
    """Mean atmospheric pressure at station elevation (Eqs. 3 & 34)

    Parameters
    ----------
    elev : ee.Image or ee.Number
        Elevation [m].
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.

    Returns
    -------
    pair : ee.Image or ee.Number
        Air pressure [kPa].

    Notes
    -----
    The current calculation in Ref-ET:
        101.3 * (((293 - 0.0065 * elev) / 293) ** (9.8 / (0.0065 * 286.9)))
    Equation 3 in ASCE-EWRI 2005:
        101.3 * (((293 - 0.0065 * elev) / 293) ** 5.26)
    Per Dr. Allen, the calculation with full precision:
        101.3 * (((293.15 - 0.0065 * elev) / 293.15) ** (9.80665 / (0.0065 * 286.9)))

    """
    if method == 'asce':
        return elev.multiply(-0.0065).add(293).divide(293).pow(5.26).multiply(101.3)
    elif method == 'refet':
        return (
            elev.multiply(-0.0065).add(293).divide(293)
            .pow(9.8 / (0.0065 * 286.9)).multiply(101.3)
        )


def sat_vapor_pressure(temperature):
    """Saturation vapor pressure from temperature (Eq. 7)

    Parameters
    ----------
    temperature : ee.Image
        Air temperature [C].

    Returns
    -------
    e : ee.Image
        Saturation vapor pressure [kPa].

    Notes
    -----
    es = 0.6108 * exp(17.27 * temperature / (temperature + 237.3))

    """
    return (
        temperature.add(237.3).pow(-1).multiply(temperature)
        .multiply(17.27).exp().multiply(0.6108)
    )


def es_slope(tmean, method='asce'):
    """Slope of the saturation vapor pressure-temperature curve (Eq. 5)

    Parameters
    ----------
    tmean : ee.Image or ee.Number
        Mean air temperature [C].
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.

    Returns
    -------
    ee.Image or ee.Number

    Notes
    -----
    4098 * 0.6108 * exp(17.27 * T / (T + 237.3)) / ((T + 237.3) ** 2))

    """
    if method == 'refet':
        return sat_vapor_pressure(tmean).multiply(4098.0).divide(tmean.add(237.3).pow(2))
    elif method == 'asce':
        return (
            tmean.add(237.3).pow(-1).multiply(tmean).multiply(17.27).exp()
            .multiply(2503.0).divide(tmean.add(237.3).pow(2))
        )


def actual_vapor_pressure(q, pair):
    """"Actual vapor pressure from specific humidity

    Parameters
    ----------
    q : ee.Image or ee.Number
        Specific humidity [kg/kg].
    pair : ee.Image or ee.Number
        Air pressure [kPa].

    Returns
    -------
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].

    Notes
    -----
    ea = q * pair / (0.622 + 0.378 * q)

    """
    return q.multiply(0.378).add(0.622).pow(-1).multiply(q).multiply(pair)


def specific_humidity(ea, pair):
    """"Specific humidity from actual vapor pressure

    Parameters
    ----------
    ea : ee.Image or ee.Number
        Specific humidity [kPa].
    pair : ee.Image or ee.Number
        Air pressure [kPa].

    Returns
    -------
    q : ee.Image or ee.Number
        Specific humidity [kg/kg].

    Notes
    -----
    q = 0.622 * ea / (pair - 0.378 * ea)

    """
    return ea.multiply(-0.378).add(pair).pow(-1).multiply(ea).multiply(0.622)


def vpd(es, ea):
    """Vapor pressure deficit

    Parameters
    ----------
    es : ee.Image or ee.Number
        Saturated vapor pressure [kPa].
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].

    Returns
    -------
    ee.Image or ee.Number
        Vapor pressure deficit [kPa].

    """
    return es.subtract(ea).max(0)


def precipitable_water(ea, pair):
    """Precipitable water in the atmosphere (Eq. D.3)

    Parameters
    ----------
    ea : ee.Image
        Vapor pressure [kPa].
    pair : ee.Image or ee.Number
        Air pressure [kPa].

    Returns
    -------
    ee.Image or ee.Number
        Precipitable water [mm].

    Notes
    -----
    w = pair * 0.14 * ea + 2.1

    """
    return ea.multiply(pair).multiply(0.14).add(2.1)


def doy_fraction(doy):
    """Fraction of the DOY in the year (Eq. 50)

    Parameters
    ----------
    doy : ee.Image or ee.Number
        Day of year.

    Returns
    -------
    ee.Image or ee.Number
        DOY fraction [radians].

    """
    return doy.multiply(2.0 * math.pi / 365)


def declination(doy, method='asce'):
    """Earth declination (Eq. 51)

    Parameters
    ----------
    doy : ee.Image or ee.Number
        Day of year.
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.

    Returns
    -------
    delta : ee.Image or ee.Number
        Earth declination [radians].

    Notes
    -----
    Original equation in Duffie & Beckman (1980) (with conversions to radians):
        delta = 23.45 * (pi / 180) * sin(2 * pi * (doy + 284) / 365)
    Equation 24 in ASCE-EWRI (2005):
        delta = 0.409 * sin((2 * pi * doy / 365) - 1.39)

    """
    if method == 'asce':
        delta = doy_fraction(doy).subtract(1.39).sin().multiply(0.409)
    else:
        delta = doy.add(284).multiply(2 * math.pi / 365).sin().multiply(23.45 * (math.pi / 180))

    return delta


def dr(doy):
    """Inverse square of the Earth-Sun Distance (Eq. 50)

    Parameters
    ----------
    doy : ee.Image or ee.Number
        Day of year.

    Returns
    -------
    ee.Image or ee.Number

    Notes
    -----
    This function returns 1 / d^2, not d, for direct use in radiance to
      TOA reflectance calculation
    dr = pi * L * d^2 / (ESUN * cos(theta)) -> pi * L / (ESUN * cos(theta) * d)

    """
    return doy_fraction(doy).cos().multiply(0.033).add(1.0)


def seasonal_correction(doy):
    """Seasonal correction for solar time (Eqs. 57 & 58)

    Parameters
    ----------
    doy : ee.Image or ee.Number
        Day of year.

    Returns
    ------
    ee.Image or ee.Number
        Seasonal correction [hour]

    Notes
    -----
    sc = 0.1645 * sin(2 * b) - 0.1255 * cos(b) - 0.0250 * sin(b)

    """
    b = doy.subtract(81).divide(364.0).multiply(2 * math.pi)
    return (
        b.multiply(2).sin().multiply(0.1645)
        .subtract(b.cos().multiply(0.1255)).subtract(b.sin().multiply(0.0250))
    )


def solar_time_rad(lon, time_mid, sc):
    """Solar time (i.e. noon is 0) (Eq. 55)

    Parameters
    ----------
    lon : ee.Image or ee.Number
        Longitude [radians].
    time_mid : ee.Image or ee.Number
        UTC time at midpoint of period [hours].
    sc : ee.Image or ee.Number
        Seasonal correction [hours].

    Returns
    -------
    ee.Image or ee.Number
        Solar time [hours].

    Notes
    -----
    This function could be integrated into the omega() function since they are
    always called together (i.e. omega(_solar_time_rad()).  It was built
    independently from omega to eventually support having a separate
    solar_time functions for longitude in degrees.
    time = (lon * 24 / (2 * math.pi)) + time_mid + sc - 12

    """
    return lon.multiply(24 / (2 * math.pi)).add(time_mid).add(sc).subtract(12)


def sha(solar_time):
    """Solar hour angle (Eq. 55)

    Parameters
    ----------
    solar_time : ee.Image or ee.Number
        Solar time (i.e. noon is 0) [hours].

    Returns
    -------
    omega : ee.Image or ee.Number
        Hour angle [radians].

    """
    omega = solar_time.multiply(2 * math.pi / 24.0)

    # Need to adjust omega so that the values go from -pi to pi
    # Values outside this range are wrapped (i.e. -3*pi/2 -> pi/2)
    return wrap(omega, -math.pi, math.pi)


def wrap(x, x_min, x_max):
    """Wrap floating point values into range

    Parameters
    ----------
    x : ee.Image or ee.Number
        Values to wrap.
    x_min : float
        Minimum value in output range.
    x_max : float
        Maximum value in output range.

    Returns
    -------
    ee.Image or ee.Number

    Notes
    -----
    This formula is used to mimic the Python modulo operator.
    Javascript/EE mod operator has the same sign as the dividend,
        so negative values stay negative.
    Python mod operator has the same sign as the divisor,
        so negative values wrap to positive.

    """
    x_range = x_max - x_min

    return x.subtract(x_min).mod(x_range).add(x_range).mod(x_range).add(x_min)


def sha_sunset(lat, delta):
    """Sunset hour angle (Eq. 59)

    Parameters
    ----------
    lat : ee.Image or ee.Number
        Latitude [radians].
    delta : ee.Image or ee.Number
        Earth declination [radians].

    Returns
    -------
    ee.Image or ee.Number
        Sunset hour angle [radians].

    Notes
    -----
    acos(-tan(lat) * tan(delta))

    """
    return lat.tan().multiply(-1).multiply(delta.tan()).acos()


def ra_daily(lat, doy, method='asce'):
    """Daily extraterrestrial radiation (Eq. 21)

    Parameters
    ----------
    lat : ee.Image or ee.Number
        latitude [radians].
    doy : ee.Image or ee.Number
        Day of year.
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.

    Returns
    -------
    ra : ee.Image or ee.Number
        Daily extraterrestrial radiation [MJ m-2 d-1].

    Notes
    -----
    Equation in ASCE-EWRI 2005 uses a solar constant of ~1366.666... W m-2
    Equation in Duffie & Beckman (?) uses a solar constant of 1367 W m-2

    """
    delta = declination(doy, method)
    omegas = sha_sunset(lat, delta)
    theta = (
        omegas.multiply(lat.sin()).multiply(delta.sin())
        .add(lat.cos().multiply(delta.cos()).multiply(omegas.sin()))
    )

    if method == 'asce':
        # (24. / math.pi) * 4.92 * dr(doy) * theta
        ra = theta.multiply(dr(doy)).multiply((24. / math.pi) * 4.92)
    else:
        # ra = (24. / math.pi) * (1367 * 0.0036) * dr(doy) * theta
        ra = theta.multiply(dr(doy)).multiply((24. / math.pi) * (1367 * 0.0036))

    return ra


def ra_hourly(lat, lon, doy, time_mid, method='asce'):
    """Hourly extraterrestrial radiation (Eq. 48)

    Parameters
    ----------
    lat : ee.Image or ee.Number
        Latitude [radians].
    lon : ee.Image or ee.Number
        Longitude [radians].
    doy : ee.Image or ee.Number
        Day of year.
    time_mid : ee.Image or ee.Number
        UTC time at midpoint of period [hours].
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.

    Returns
    -------
    ra : ee.Image or ee.Number
        Hourly extraterrestrial radiation [MJ m-2 h-1].

    Notes
    -----
    Equation in ASCE-EWRI 2005 uses a solar constant of ~1366.666... W m-2
    Equation in Duffie & Beckman (?) uses a solar constant of 1367 W m-2

    """
    omega = sha(solar_time_rad(lon, time_mid, seasonal_correction(doy)))
    delta = declination(doy, method)
    omegas = sha_sunset(lat, delta)

    # Solar time as start and end of period (Eqs. 53 & 54)
    # Modify omega1 and omega2 at sunrise and sunset (Eq. 56)
    omega1 = omega.subtract(math.pi / 24).max(omegas.multiply(-1)).min(omegas)
    omega2 = omega.add(math.pi / 24).max(omegas.multiply(-1)).min(omegas)
    omega1 = omega1.min(omega2)

    # Extraterrestrial radiation (Eq. 48)
    theta = (
        omega2.subtract(omega1).multiply(lat.sin()).multiply(delta.sin())
        .add(lat.cos().multiply(delta.cos()).multiply(omega2.sin().subtract(omega1.sin())))
    )
    if method == 'asce':
        ra = theta.multiply(dr(doy)).multiply((12. / math.pi) * 4.92)
    else:
        ra = theta.multiply(dr(doy)).multiply((12. / math.pi) * (1367 * 0.0036))

    return ra


def rso_daily(ea, pair, ra, doy, lat):
    """Full daily clear sky solar radiation formulation (Appendix D)

    Parameters
    ----------
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].
    pair : ee.Image or ee.Number
        Air pressure [kPa].
    ra : ee.Image or ee.Number
        Extraterrestrial radiation [MJ m-2 d-1].
    doy : ee.Image or ee.Number
        Day of year.
    lat : ee.Image or ee.Number
        Latitude [rad].

    Returns
    -------
    rso : ee.Image or ee.Number
        Daily clear sky solar radiation [MJ m-2 d-1].
        Output data type will match "ea" data type.

    """
    # Sin() of the angle of the sun above the horizon (D.5 and Eq. 62)
    # Compute from latitude since doy is more likely to be an ee.Number
    sin_beta_24 = (
        lat.multiply(doy_fraction(doy).subtract(1.39).sin()).multiply(0.3)
        .add(0.85).subtract(lat.pow(2).multiply(0.42)).sin().max(0.1)
    )

    # Precipitable water
    w = precipitable_water(ea, pair)

    # Clearness index for direct beam radiation (Eq. D.2)
    # Limit sin_beta >= 0.01 so that KB does not go undefined
    kb = (
        w.divide(sin_beta_24).pow(0.4).multiply(-0.075)
        .add(pair.multiply(-0.00146).divide(sin_beta_24)).exp().multiply(0.98)
    )

    # Transmissivity index for diffuse radiation (Eq. D.4)
    kd = kb.multiply(-0.36).add(0.35).min(kb.multiply(0.82).add(0.18))

    return kb.add(kd).multiply(ra)


def rso_hourly(ea, ra, pair, doy, time_mid, lat, lon, method='asce'):
    """Full hourly clear sky solar radiation formulation (Appendix D)

    Parameters
    ----------
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].
    ra : ee.Image or ee.Number
        Extraterrestrial radiation [MJ m-2 h-1].
    pair : ee.Image or ee.Number
        Air pressure [kPa].
    doy : ee.Image or ee.Number
        Day of year.
    time_mid : ee.Image or ee.Number
        UTC time at midpoint of period [hours].
    lat : ee.Image or ee.Number
        Latitude [rad].
    lon : ee.Image or ee.Number
        Longitude [rad].
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.
        Passed through to declination function.

    Returns
    -------
    rso : ee.Image or ee.Number
        Hourly clear sky solar radiation [MJ m-2 h-1].
        Output data type will match "ra" data type.

    """
    sc = seasonal_correction(doy)
    omega = sha(solar_time_rad(lon, time_mid, sc))

    # sin of the angle of the sun above the horizon (D.6 and Eq. 62)
    delta = declination(doy, method)
    sin_beta = (
        lat.sin().multiply(delta.sin())
        .add(lat.cos().multiply(delta.cos()).multiply(omega.cos()))
    )

    # Precipitable water
    w = precipitable_water(ea, pair)

    # Clearness index for direct beam radiation (Eq. D.2)
    # Limit sin_beta >= 0.01 so that KB does not go undefined
    kt = 1.0
    kb = (
        w.divide(sin_beta.max(0.01)).pow(0.4).multiply(-0.075)
        .add(pair.multiply(-0.00146).divide(sin_beta.max(0.01).multiply(kt)))
        .exp().multiply(0.98)
    )

    # Transmissivity index for diffuse radiation (Eq. D.4)
    kd = kb.multiply(-0.36).add(0.35).min(kb.multiply(0.82).add(0.18))

    return kb.add(kd).multiply(ra)


def rso_simple(ra, elev):
    """Simplified daily/hourly clear sky solar formulation (Eqs. 19 & 45)

    Parameters
    ----------
    ra : ee.Image or ee.Number
        Extraterrestrial radiation [MJ m-2 d-1 or MJ m-2 h-1].
    elev : ee.Image or ee.Number
        Elevation [m].

    Returns
    -------
    rso : ee.Image or ee.Number
        Clear sky solar radiation [MJ m-2 d-1 or MJ m-2 h-1].
        Output data type will match "ra" data type.

    Notes
    -----
    rso = (0.75 + 2E-5 * elev) * ra

    """
    return ra.multiply(elev.multiply(2E-5).add(0.75))


def fcd_daily(rs, rso):
    """Daytime cloudiness fraction (Eq. 18)

    Parameters
    ----------
    rs : ee.Image or ee.Number
        Measured solar radiation [MJ m-2 d-1].
    rso : ee.Image or ee.Number
        Clear sky solar radiation [MJ m-2 d-1].

    Returns
    -------
    fcd : ee.Image or ee.Number
        Output data type will match "rs" data type.

    Notes
    -----
    fcd = 1.35 * min(max(rs / rso, 0.3), 1.0) - 0.35

    """
    return rs.divide(rso).max(0.3).min(1.0).multiply(1.35).subtract(0.35)


def fcd_hourly(rs, rso, doy, time_mid, lat, lon, method='asce'):
    """Cloudiness fraction (Eq. 45)

    Parameters
    ----------
    rs : ee.Image or ee.Number
        Measured solar radiation [MJ m-2 h-1].
    rso : ee.Image or ee.Number
        Clear sky solar radiation [MJ m-2 h-1].
    doy : ee.Image or ee.Number
        Day of year.
    time_mid : ee.Image or ee.Number
        UTC time at midpoint of period [hours].
    lat : ee.Image or ee.Number
        Latitude [rad].
    lon : ee.Image or ee.Number
        Longitude [rad].
    method : {'asce' (default), 'refet'}, optional
        Calculation method:
        * 'asce' -- Calculations will follow ASCE-EWRI 2005 [1] equations.
        * 'refet' -- Calculations will follow RefET software.
        Passed through to declination function.

    Returns
    -------
    fcd : ee.Image or ee.Number
        Output data type will match "rs" data type.

    """
    # DEADBEEF - These values are only needed for identifying low sun angles
    sc = seasonal_correction(doy)
    delta = declination(doy, method)
    omega = sha(solar_time_rad(lon, time_mid, sc))
    beta = (
        lat.sin().multiply(delta.sin())
        .add(lat.cos().multiply(delta.cos()).multiply(omega.cos())).asin()
    )

    fcd = rs.divide(rso).max(0.3).min(1).multiply(1.35).subtract(0.35)

    # Intentionally not using where() so that function will work with ee.Number
    fcd = fcd.max(beta.lt(0.3))
    # fcd = fcd.where(beta.lt(0.3), 1)

    # # DEADBEEF - Code from NumPy functions
    # fcd = np.ones(rso.shape)
    # fcd[rso > 0] = 1.35 * np.clip(rs[rso > 0] / rso[rso > 0], 0.3, 1) - 0.35
    #
    # # For now just set fcd to 1 for low sun angles
    # # DEADBEEF - Still need to get daytime value of fcd when beta > 0.3
    # #   Get closest value in time (array space) when beta > 0.3
    # fcd[beta < 0.3] = 1

    return fcd


def rnl_daily(tmax, tmin, ea, fcd):
    """Daily net long-wave radiation  (Eq. 17)

    Parameters
    ----------
    tmax : ee.Image or ee.Number
        Daily maximum air temperature [C].
    tmin : ee.Image or ee.Number
        Daily minimum air temperature [C].
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].
    fcd : ee.Image or ee.Number
        cloudiness fraction.

    Returns
    -------
    ee.Image or ee.Number
        Daily net long-wave radiation [MJ m-2 d-1].
        Output data type will match "tmax" data type.

    Notes
    -----
    rnl = 4.901E-9 * fcd * (0.34 - 0.14 * sqrt(ea)) *
          0.5 * ((tmax + 273.16) ** 4 + (tmin + 273.16) ** 4))

    """
    return (
        tmax.add(273.16).pow(4).add(tmin.add(273.16).pow(4)).multiply(0.5)
        .multiply(ea.sqrt().multiply(-0.14).add(0.34))
        .multiply(fcd).multiply(4.901E-9)
    )


def rnl_hourly(tmean, ea, fcd):
    """Hourly net long-wave radiation  (Eq. 44)

    Parameters
    ----------
    tmean : ee.Image or ee.Number
        Mean hourly air temperature [C].
    ea : ee.Image or ee.Number
        Actual vapor pressure [kPa].
    fcd : ee.Image or ee.Number
        Cloudiness fraction.

    Returns
    -------
    ee.Image or ee.Number
        Hourly net long-wave radiation [MJ m-2 h-1].
        Output data type will match "tmean" data type.

    Notes
    -----
    rnl = 2.042E-10 * fcd * (0.34 - 0.14 * sqrt(ea)) * ((tmean + 273.16) ** 4)

    """
    return (
        tmean.add(273.16).pow(4)
        .multiply(ea.sqrt().multiply(-0.14).add(0.34))
        .multiply(fcd).multiply(2.042E-10)
    )


def rn(rs, rnl):
    """Net daily/hourly radiation (Eqs. 15 & 16)

    Parameters
    ----------
    rs : ee.Image or ee.Number
        Measured solar radiation [MJ m-2 d-1 or MJ m-2 h-1].
    rnl : ee.Image or ee.Number
        Hourly net long-wave radiation [MJ m-2 d-1 or MJ m-2 h-1].

    Returns
    -------
    ee.Image or ee.Number
        Hourly net long-wave radiation [MJ m-2 d-1 or MJ m-2 h-1].
        Output data type will match "rnl" data type.

    Notes
    -----
    rnl = 0.77 * rs - rnl

    """
    # Calculation is intentionally being made from Rnl (which is computed from temperature)
    #   to keep a reference to the source image
    return rnl.multiply(-1).add(rs.multiply(0.77))


def wind_height_adjust(uz, zw):
    """Wind speed at 2 m height based on full logarithmic profile (Eq. 33)

    Parameters
    ----------
    uz : ee.Image or ee.Number
        Wind speed at measurement height [m s-1].
    zw : ee.Image or ee.Number
        Wind measurement height [m].

    Returns
    -------
    ee.Image or ee.Number
        Wind speed at 2 m height [m s-1].

    Notes
    -----
    u2 = uz * 4.87 / log(67.8 * zw - 5.42)

    """
    return uz.multiply(4.87).divide(zw.multiply(67.8).subtract(5.42).log())



import ee
import pytest

from openet.refetgee import Hourly
import openet.refetgee.units as units
import utils

# Eventually move to conftest.py or a separate file
# Fallon AgriMet site parameters
s_args = {
    'elev': 1208.5,
    'lat': 39.4575,
    'lon': -118.77388,
    'pair': 87.81876435813037,
    'pair_asce': 87.80710537212929,
    'zw': 3.0,
}

# Hourly test parameters for 2015-07-01 18:00 UTC (11:00 AM PDT)
# DEADBEEF - The eto_refet value changes...
h_args = {
    'doy': 182,
    'ea': 1.1990099614301906,
    'es': 5.09318785259078,
    'eto_refet': 0.6068613650177562,
    'eto_asce': 0.6063515410076268,
    'etr_refet': 0.7201865213918281,
    'etr_asce': 0.7196369609713682,
    'q': 0.008536365803069757,          # Computed from Ea from Tdew
    'q_asce': 0.00853750513849305,      # Computed from Ea from Tdew
    'ra': 4.30824147948541,
    'rnl': 0.22897874401150786,
    'rs': 61.16 * 0.041868,             # Conversion from Langleys to MJ m-2
    'tdew': units.f2c(49.36),
    'time': 18.0,
    'time_mid': 18.5,
    'tmean': units.f2c(91.80),
    'uz': 3.33 * 0.44704,               # Conversion from mph to m s-1
    'u2': 1.3709275319197722,
}

# Compute wind x and y components
h_args['uz_x'] = h_args['uz'] / (2 ** 0.5)
h_args['uz_y'] = h_args['uz'] / (2 ** 0.5)
# h_args['uz_x'] = h_args['uz']
# h_args['uz_y'] = 0.0


# Test full hourly functions with positional inputs
def test_refet_hourly_input_positions():
    refet = Hourly(
        ee.Image.constant(h_args['tmean']), ee.Image.constant(h_args['ea']),
        ee.Image.constant(h_args['rs']), ee.Image.constant(h_args['uz']),
        ee.Image.constant(s_args['zw']), ee.Number(s_args['elev']),
        ee.Number(s_args['lat']), ee.Number(s_args['lon']),
        ee.Number(h_args['doy']), ee.Number(h_args['time']), method='refet'
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_refet'])


# Test full hourly calculations with keyword inputs
def test_refet_hourly_default_method_etr():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']),
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_asce_method_etr():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Number(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_refet_method_etr():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']), method='refet',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_refet'])


def test_refet_hourly_default_method_eto():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']),
    )

    output = utils.constant_image_value(refet.eto)

    assert float(output['eto']) == pytest.approx(h_args['eto_asce'])


def test_refet_hourly_asce_method_eto():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']), method='asce',
    )

    output = utils.constant_image_value(refet.eto)

    assert float(output['eto']) == pytest.approx(h_args['eto_asce'])


def test_refet_hourly_asce_method_eto():
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']), method='refet',
    )

    output = utils.constant_image_value(refet.eto)

    assert float(output['eto']) == pytest.approx(h_args['eto_refet'])


@pytest.mark.parametrize(
    'surface, expected',
    [
        ['etr', h_args['etr_refet']],
        ['alfalfa', h_args['etr_refet']],
        ['tall', h_args['etr_refet']],
        ['eto', h_args['eto_refet']],
        ['grass', h_args['eto_refet']],
        ['short', h_args['eto_refet']],
    ]
)
def test_refet_daily_etsz(surface, expected):
    refet = Hourly(
        tmean=ee.Image.constant(h_args['tmean']), ea=ee.Image.constant(h_args['ea']),
        rs=ee.Image.constant(h_args['rs']), uz=ee.Image.constant(h_args['uz']),
        zw=ee.Image.constant(s_args['zw']), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        doy=ee.Number(h_args['doy']), time=ee.Number(h_args['time']), method='refet',
    )

    output = utils.constant_image_value(refet.etsz(surface).rename(['etsz']))

    assert float(output['etsz']) == pytest.approx(expected)


def test_refet_hourly_nldas_etr():
    """Generate a mock NLDAS image from the test values"""
    time_start = ee.Date(f'2015-07-01T{int(h_args["time"])}:00:00', 'UTC').millis()

    input_img = (
        ee.Image.constant([
            h_args['tmean'], h_args['q_asce'], h_args['rs'] / 0.0036,
            h_args['uz_x'], h_args['uz_y']
        ])
        .rename(['temperature', 'specific_humidity', 'shortwave_radiation', 'wind_u', 'wind_v'])
        .set('system:time_start', time_start)
    )

    refet = Hourly.nldas(
        ee.Image(input_img), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_rtma_etr():
    """Generate a mock RTMA image from the test values"""
    time_start = ee.Date(f'2015-07-01T{int(h_args["time"])}:00:00', 'UTC').millis()

    input_img = (
        ee.Image.constant([h_args['tmean'], h_args['q_asce'], h_args['uz']])
        .rename(['TMP', 'SPFH', 'WIND'])
        .set('system:time_start', time_start)
    )

    refet = Hourly.rtma(
        ee.Image(input_img), rs=ee.Image.constant(h_args['rs']),
        elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_era5_etr():
    """Generate a mock ERA5 image from the test values"""
    time_start = ee.Date(f'2015-07-01T{int(h_args["time"])}:00:00', 'UTC').millis()

    input_img = (
        ee.Image.constant([
            h_args['tmean'] + 273.15, h_args['tdew'] + 273.15,
            h_args['rs'] * 1000000, h_args['uz_x'], h_args['uz_y']
        ])
        .rename(['temperature_2m', 'dewpoint_temperature_2m',
                 'surface_solar_radiation_downwards',
                 'u_component_of_wind_10m', 'v_component_of_wind_10m'])
        .set('system:time_start', time_start)
    )

    refet = Hourly.era5(
        ee.Image(input_img), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_era5_land_etr():
    """Generate a mock ERA5-Land image from the test values"""
    time_start = ee.Date(f'2015-07-01T{int(h_args["time"])}:00:00', 'UTC').millis()

    input_img = (
        ee.Image.constant([
            h_args['tmean'] + 273.15, h_args['tdew'] + 273.15,
            h_args['rs'] * 1000000, h_args['uz_x'], h_args['uz_y']
        ])
        .rename(['temperature_2m', 'dewpoint_temperature_2m',
                 'surface_solar_radiation_downwards_hourly',
                 'u_component_of_wind_10m', 'v_component_of_wind_10m'])
        .set('system:time_start', time_start)
    )

    refet = Hourly.era5_land(
        ee.Image(input_img), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), lon=ee.Number(s_args['lon']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(h_args['etr_asce'])


def test_refet_hourly_era5_land_fill_edge_cells():
    """Check that the fill_edge_cells flag works for an edge cell along the coast of England"""
    input_img = ee.Image('ECMWF/ERA5_LAND/HOURLY/20150701T20')
    output = utils.point_image_value(Hourly.era5_land(input_img, fill_edge_cells=False).etr, xy=[0.0, 50.7])
    assert output['etr'] is None
    output = utils.point_image_value(Hourly.era5_land(input_img, fill_edge_cells=True).etr, xy=[0.0, 50.7])
    assert output['etr'] is not None


# TODO: Add a test for using the default Rs when one is not provided



import datetime as dt
import os

import ee
import numpy as np
import pandas as pd
import pytest

from openet.refetgee import Daily
import openet.refetgee.units as units
import utils


# Test daily functions using actual RefET input/output files
class DailyData():
    """Setup daily validation data from Fallon AgriMet station"""
    val_ws = os.path.join(os.getcwd(), 'openet', 'refetgee', 'tests', 'data')

    csv_path = os.path.join(val_ws, 'FALN_Agrimet_daily_raw_2015.csv')
    out_path = os.path.join(val_ws, 'FALN_Agrimet_daily_raw_2015.out')

    # Read in the inputs CSV file using pandas
    csv_df = pd.read_csv(csv_path, engine='python', na_values='NO RECORD')
    csv_df.rename(
        columns={'MN': 'TMIN', 'MX': 'TMAX', 'YM': 'TDEW', 'UA': 'WIND', 'SR': 'RS'},
        inplace=True,
    )
    csv_df['DATE'] = csv_df[['YEAR', 'MONTH', 'DAY']].apply(
        lambda x: dt.datetime(*x).strftime('%Y-%m-%d'), axis=1
    )
    csv_df.set_index('DATE', inplace=True, drop=True)

    # Convert inputs units
    csv_df['TMIN'] = units.f2c(csv_df['TMIN'])
    csv_df['TMAX'] = units.f2c(csv_df['TMAX'])
    csv_df['TDEW'] = units.f2c(csv_df['TDEW'])
    csv_df['WIND'] *= 0.44704
    csv_df['RS'] *= 0.041868  # Conversion from Langleys to MJ m-2 to match RefET
    # csv_df['RS'] *= 0.041840  # Alternate conversion from Langleys to MJ m-2
    csv_df['EA'] = 0.6108 * np.exp(17.27 * csv_df['TDEW'] / (csv_df['TDEW'] + 237.3))

    # Eventually compare ancillary functions directly to IN2 values
    # # Identify the row number of the IN2 data
    # with open(in2_path) as in2_f:
    #     in2_data = in2_f.readlines()
    # in2_start = [i for i, x in enumerate(in2_data) if x.startswith(' Mo Da Year ')][0]
    # # Read in the IN2 file using pandas
    # in2_df = pd.read_table(in2_path, sep='\s+', skiprows=in2_start, header=[0, 1, 2])
    # in2_df.rename(
    #     columns={'Year': 'YEAR', 'Mo': 'MONTH', 'Da': 'DAY', 'DoY': 'DOY'}, inplace=True
    # )
    # in2_df['DATE'] = in2_df[['YEAR', 'MONTH', 'DAY']].apply(
    #     lambda x: dt.datetime(*x).strftime('%Y-%m-%d'), axis=1
    # )
    # in2_df.set_index('DATE', inplace=True, drop=True)

    # Identify the row number of the OUT data
    with open(out_path) as out_f:
        out_data = out_f.readlines()
    out_start = [i for i, x in enumerate(out_data) if x.startswith(' Mo Day Yr')][0]
    # Read in the OUT file using pandas (skip header and units)
    out_df = pd.read_table(
        out_path, sep='\\s+', index_col=False,
        skiprows=list(range(out_start)) + [out_start + 1],
    )
    out_df.rename(
        columns={
            'Yr': 'YEAR', 'Mo': 'MONTH', 'Day': 'DAY', 'Tmax': 'TMAX', 'Tmin': 'TMIN',
            'Wind': 'WIND', 'Rs': 'RS', 'DewP': 'TDEW',
        },
        inplace=True,
    )
    out_df['DATE'] = out_df[['YEAR', 'MONTH', 'DAY']].apply(
        lambda x: dt.datetime(*x).strftime('%Y-%m-%d'), axis=1
    )
    out_df.set_index('DATE', inplace=True, drop=True)

    # Read the station properties from the IN2 file for now
    # The values should probably be read using a regular expression
    for line in out_data:
        if line.strip().startswith('The anemometer height is'):
            zw = float(line.split(':')[1].split()[0])
        elif line.strip().startswith('The weather station elevation is'):
            elev = float(line.split(':')[1].split()[0])
        elif line.strip().startswith('The weather station latitude is'):
            lat = float(line.split(':')[1].split()[0])

    # Get list of date strings from the input CSV file
    values, ids = [], []
    for test_date in list(csv_df.index):
        # Datetime that has issues with fcd calculation
        if not test_date.startswith('2015-07'):
            continue

        test_dt = dt.datetime.strptime(test_date, '%Y-%m-%d')
        # Can the surface type be parameterized inside pytest_generate_tests?
        for surface in ['ETr', 'ETo']:
            date_values = (
                csv_df
                .loc[test_date, ['TMIN', 'TMAX', 'EA', 'RS', 'WIND']]
                .rename({'TMIN': 'tmin', 'TMAX': 'tmax', 'EA': 'ea', 'RS': 'rs', 'WIND': 'uz'})
                .to_dict()
            )
            date_values.update({
                'surface': surface.lower(),
                'expected': out_df.loc[test_date, surface],
                'doy': int(dt.datetime.strptime(test_date, '%Y-%m-%d').strftime('%j')),
                'zw': zw,
                'elev': elev,
                'lat': lat,
                'rso_type': 'full',
            })
            values.append(date_values)
            ids.append('{}-{}'.format(test_date, surface))


def pytest_generate_tests(metafunc):
    if 'daily_params' not in metafunc.fixturenames:
        return
    daily = DailyData()
    metafunc.parametrize('daily_params', daily.values, ids=daily.ids, scope='module')


def test_refet_daily_output(daily_params):
    """Test daily RefET calculation at a single point and time"""
    # If I don't copy, the pop changes the test values in daily_data()
    inputs = daily_params.copy()
    surface = inputs.pop('surface')
    expected = inputs.pop('expected')

    # ETr/ETo values only have 4 significant figures
    # Small number of days don't match if difference is set < 0.008
    diff = 0.05 if expected >= 10.0 else 0.008

    # Set primary input variables as constant images
    adj_inputs = {}
    for k, v in inputs.items():
        if k == 'rso_type':
            adj_inputs[k] = v
        elif k in ['tmin', 'tmax', 'ea', 'rs', 'uz']:
            adj_inputs[k] = ee.Image.constant(v)
        else:
            adj_inputs[k] = ee.Number(v)

    if surface.lower() == 'etr':
        refet = Daily(**adj_inputs).etr
    elif surface.lower() == 'eto':
        refet = Daily(**adj_inputs).eto
    output = utils.constant_image_value(refet)

    assert float(output[surface.lower()]) == pytest.approx(expected, abs=diff)



import logging
import time

import ee


def get_info(ee_obj, max_retries=4):
    """Make an exponential back off getInfo call on an Earth Engine object"""
    # output = ee_obj.getInfo()
    output = None
    for i in range(1, max_retries):
        try:
            output = ee_obj.getInfo()
        except ee.ee_exception.EEException as e:
            if ('Earth Engine memory capacity exceeded' in str(e) or
                    'Earth Engine capacity exceeded' in str(e) or
                    'Too many concurrent aggregations' in str(e) or
                    'Computation timed out.' in str(e)):
                # TODO: Maybe add 'Connection reset by peer'
                logging.info(f'    Resending query ({i}/{max_retries})')
                logging.info(f'    {e}')
            else:
                # TODO: What should happen for unexpected EE exceptions?
                #   It might be better to reraise the exception and exit
                logging.info(f'    {e}')
                logging.info('    Unhandled Earth Engine exception')
                continue
        except Exception as e:
            logging.info(f'    Resending query ({i}/{max_retries})')
            logging.debug(f'    {e}')

        if output is not None:
            break

        time.sleep(i ** 3)

    return output


def constant_image_value(image, crs='EPSG:32613', scale=1):
    """Extract the output value from a "constant" image"""
    rr_params = {
        'reducer': ee.Reducer.first(),
        'geometry': ee.Geometry.Rectangle([0, 0, 10, 10], crs, False),
        'scale': scale,
    }
    return get_info(ee.Image(image).reduceRegion(**rr_params))


def point_image_value(image, xy, scale=1):
    """Extract the output value from a calculation at a point"""
    rr_params = {
        'reducer': ee.Reducer.first(),
        'geometry': ee.Geometry.Point(xy),
        'scale': scale,
    }
    return get_info(ee.Image(image).reduceRegion(**rr_params))



import math

import pytest

import openet.refetgee.units as units


def test_deg2rad(d=30, r=(math.pi / 6)):
    assert units.deg2rad(d) == pytest.approx(r)


def test_rad2deg(r=(math.pi / 6), d=30):
    assert units.rad2deg(r) == pytest.approx(d)


def test_c2f(c=20, f=68):
    assert units.c2f(c) == f


def test_f2c(f=68, c=20):
    assert units.f2c(f) == c



import ee
import pytest

import openet.refetgee.calcs as calcs
import openet.refetgee.units as units
import utils

# Eventually move to conftest.py or a separate file
# Fallon AgriMet site parameters
s_args = {
    'elev': 1208.5,
    'lat': units.deg2rad(39.4575),
    'lon': units.deg2rad(-118.77388),
    'pair': 87.81876435813037,
    'pair_asce': 87.80710537212929,
    'zw': 3.0,
}
# Daily test parameters for 2015-07-01
d_args = {
    'doy': 182,
    'delta': 0.40352881013673136,
    'delta_asce': 0.4029517192078854,
    'doy_frac': 3.132985550429273,
    'dr': 0.9670012223491632,
    'ea': 1.2206674169951346,
    'es': 4.6747236227258835,
    'es_slope': 0.23489129849801055,
    'es_slope_asce': 0.23488581814172638,
    # 'eto': 7.942481120179387,
    # 'etr': 10.571560006380153,
    'fcd': 0.8569860867772078,
    'omega_s': 1.9298904620748385,
    'q': 0.008691370735727117,          # Computed from Ea from Tdew
    'q_asce': 0.008692530868140688,     # Computed from Ea from Tdew
    'ra': 41.67610845067083,
    'ra_asce': 41.64824567735701,
    'rn': 15.174377350374279,
    'rnl': 6.556533974825727,
    'rs': 674.07 * 0.041868,            # Conversion from Langleys to MJ m-2
    'rso': 31.565939444861765,
    'rso_simple': 32.26439287925584,
    'sc': -0.05874166519510547,
    'tdew': units.f2c(49.84),
    'tmin': units.f2c(66.65),
    'tmax': units.f2c(102.80),
    # 'tmean': f2c(84.725),
    'w': 17.107650595384076,
    'uz': 4.80 * 0.44704,               # Conversion from mph to m s-1
    'u2': 1.976111757722194,
}
# Hourly test parameters for 2015-07-01 18:00 UTC (11:00 AM PDT)
h_args = {
    'doy': 182,
    'ea': 1.1990099614301906,           # Computed from Tdew
    'es': 5.09318785259078,
    # 'eto': 0.6065255163817055,
    # 'etr': 0.7201865213918281,
    'fcd': 0.6816142001345745,
    'fcd_asce': 0.6816142001345745,
    # 'omega': -0.3866777826605525,     # Computed from time_mid
    'omega': -0.5175774765601271,       # Computed from time to match IN2
    'q': 0.008536365803069757,          # Computed from Ea from Tdew
    'q_asce': 0.00853750513849305,      # Computed from Ea from Tdew
    'ra': 4.30824147948541,
    'ra_asce': 4.30635461533285,
    'rnl': 0.22897874401150786,
    'rs': 61.16 * 0.041868,             # Conversion from Langleys to MJ m-2
    'rso': 3.350936122776373,
    'rso_asce': 3.350851392549374,
    'rso_simple': 3.33531130617322,
    # 'solar_time': -1.4770003318617722,  # Computed from time_mid
    'solar_time': -1.9770003318617722,  # Computed from time to match IN2
    'tdew': units.f2c(49.36),
    'time': 18.0,
    'time_mid': 18.5,
    'tmean': units.f2c(91.80),
    'uz': 3.33 * 0.44704,  # Conversion from mph to m s-1
    'u2': 1.3709275319197722,
}

# # Playing around with passing a dictionary of function keyword arguments
# daily_args = {
#     k: v for k, v in daily.items()
#     if k in ['tmin', 'tmax', 'ea', 'rs', 'uz', 'zw', 'doy']}
# daily_args.update({'ref_type':'etr', 'rso_type': 'full', 'rso': None})
# hourly_args = {
#     k: v for k, v in daily.items()
#     if k in ['tmean', 'ea', 'rs', 'uz', 'zw', 'doy', 'time']}
# hourly_args.update({'ref_type':'etr'})


# Test ancillary functions with positional inputs
def test_air_pressure_default(elev=s_args['elev'], expected=s_args['pair_asce']):
    output = utils.get_info(calcs.air_pressure(elev=ee.Number(elev)))
    assert float(output) == pytest.approx(expected)

def test_air_pressure_asce(elev=s_args['elev'], expected=s_args['pair_asce']):
    output = utils.get_info(calcs.air_pressure(elev=ee.Number(elev), method='asce'))
    assert float(output) == pytest.approx(expected)

def test_air_pressure_refet(elev=s_args['elev'], expected=s_args['pair']):
    output = utils.get_info(calcs.air_pressure(elev=ee.Number(elev), method='refet'))
    assert float(output) == pytest.approx(expected)

def test_air_pressure_position(elev=s_args['elev'], expected=s_args['pair']):
    output = utils.get_info(calcs.air_pressure(ee.Number(elev), 'refet'))
    assert float(output) == pytest.approx(expected)

def test_air_pressure_image(elev=s_args['elev'], expected=s_args['pair_asce']):
    output = utils.constant_image_value(calcs.air_pressure(elev=ee.Image.constant(elev)))
    assert float(output['constant']) == pytest.approx(expected)


@pytest.mark.parametrize(
    'tdew, expected',
    [
        [d_args['tdew'], d_args['ea']],
        [h_args['tmean'], h_args['es']],
        [h_args['tdew'], h_args['ea']],
    ]
)
def test_sat_vapor_pressure_number(tdew, expected):
    output = utils.get_info(calcs.sat_vapor_pressure(ee.Number(tdew)))
    assert float(output) == pytest.approx(expected)

def test_sat_vapor_pressure_keyword(tdew=d_args['tdew'], expected=d_args['ea']):
    output = utils.get_info(calcs.sat_vapor_pressure(temperature=ee.Number(tdew)))
    assert float(output) == pytest.approx(expected)

def test_sat_vapor_pressure_image(tdew=d_args['tdew'], expected=d_args['ea']):
    output = utils.constant_image_value(calcs.sat_vapor_pressure(ee.Image.constant(tdew)))
    assert float(output['constant']) == pytest.approx(expected)


@pytest.mark.parametrize(
    'ea, pair, expected',
    [
        [d_args['ea'], s_args['pair'], d_args['q']],
        [d_args['ea'], s_args['pair_asce'], d_args['q_asce']],
        [h_args['ea'], s_args['pair'], h_args['q']],
        [h_args['ea'], s_args['pair_asce'], h_args['q_asce']],
    ]
)
def test_specific_humidity_number(ea, pair, expected):
    output = utils.get_info(calcs.specific_humidity(
        ea=ee.Number(ea), pair=ee.Number(pair)
    ))
    assert float(output) == pytest.approx(expected)

def test_specific_humidity_position(ea=d_args['ea'], pair=s_args['pair'], expected=d_args['q']):
    output = utils.get_info(calcs.specific_humidity(ee.Number(ea), ee.Number(pair)))
    assert float(output) == pytest.approx(expected)

def test_specific_humidity_image(ea=d_args['ea'], pair=s_args['pair'], expected=d_args['q']):
    output = utils.constant_image_value(calcs.specific_humidity(
        ea=ee.Image.constant(ea), pair=ee.Number(pair)
    ))
    assert float(output['constant']) == pytest.approx(expected)


def test_actual_vapor_pressure_number(q=d_args['q'], pair=s_args['pair'], expected=d_args['ea']):
    output = utils.get_info(calcs.actual_vapor_pressure(q=ee.Number(q), pair=ee.Number(pair)))
    assert float(output) == pytest.approx(expected)

def test_actual_vapor_pressure_asce(
        q=d_args['q_asce'], pair=s_args['pair_asce'], expected=d_args['ea']
):
    output = utils.get_info(calcs.actual_vapor_pressure(q=ee.Number(q), pair=ee.Number(pair)))
    assert float(output) == pytest.approx(expected)

def test_actual_vapor_pressure_position(q=d_args['q'], pair=s_args['pair'], expected=d_args['ea']):
    output = utils.get_info(calcs.actual_vapor_pressure(ee.Number(q), ee.Number(pair)))
    assert float(output) == pytest.approx(expected)

def test_actual_vapor_pressure_image(q=d_args['q'], pair=s_args['pair'], ea=d_args['ea']):
    output = utils.constant_image_value(calcs.actual_vapor_pressure(
        q=ee.Image.constant(q), pair=ee.Number(pair)
    ))
    assert float(output['constant']) == pytest.approx(ea)


def test_vpd_number(es=d_args['es'], ea=d_args['ea']):
    output = utils.get_info(calcs.vpd(es=ee.Number(es), ea=ee.Number(ea)))
    assert float(output) == pytest.approx(float(es - ea))

    # Check that negative VPD's are set to 0
    output = utils.get_info(calcs.vpd(es=ee.Number(es), ea=ee.Number(es + 1)))
    assert float(output) == pytest.approx(0)

def test_vpd_image(es=d_args['es'], ea=d_args['ea']):
    output = utils.constant_image_value(calcs.vpd(es=ee.Image.constant(es), ea=ee.Number(ea)))
    assert float(output['constant']) == pytest.approx(float(es-ea))


def test_es_slope_default(tmin=d_args['tmin'], tmax=d_args['tmax'], expected=d_args['es_slope_asce']):
    output = utils.get_info(calcs.es_slope(tmean=ee.Number(0.5 * (tmin + tmax))))
    assert float(output) == pytest.approx(expected)

def test_es_slope_asce(tmin=d_args['tmin'], tmax=d_args['tmax'], expected=d_args['es_slope_asce']):
    output = utils.get_info(calcs.es_slope(
        tmean=ee.Number(0.5 * (tmin + tmax)), method='asce'
    ))
    assert float(output) == pytest.approx(expected)

def test_es_slope_refet(tmin=d_args['tmin'], tmax=d_args['tmax'], expected=d_args['es_slope']):
    output = utils.get_info(calcs.es_slope(tmean=ee.Number(0.5 * (tmin + tmax)), method='refet'))
    assert float(output) == pytest.approx(expected)

def test_es_slope_position(tmin=d_args['tmin'], tmax=d_args['tmax'], expected=d_args['es_slope']):
    output = utils.get_info(calcs.es_slope(ee.Number(0.5 * (tmin + tmax)), 'refet'))
    assert float(output) == pytest.approx(expected)

def test_es_slope_image(tmin=d_args['tmin'], tmax=d_args['tmax'], expected=d_args['es_slope_asce']):
    output = utils.constant_image_value(calcs.es_slope(
        tmean=ee.Image.constant(0.5 * (tmin + tmax))
    ))
    assert float(output['constant']) == pytest.approx(expected)


def test_precipitable_water_number(ea=d_args['ea'], pair=s_args['pair'], expected=d_args['w']):
    output = utils.get_info(calcs.precipitable_water(ee.Number(ea), ee.Number(pair)))
    assert float(output) == pytest.approx(expected)

def test_precipitable_water_image_1(ea=d_args['ea'], pair=s_args['pair'], expected=d_args['w']):
    output = utils.constant_image_value(calcs.precipitable_water(
        ee.Image.constant(ea), ee.Number(pair)
    ))
    assert float(output['constant']) == pytest.approx(expected)

def test_precipitable_water_image_2(ea=d_args['ea'], pair=s_args['pair'], expected=d_args['w']):
    output = utils.constant_image_value(calcs.precipitable_water(
        ee.Image.constant(ea), ee.Image.constant(pair)
    ))
    assert float(output['constant']) == pytest.approx(expected)


def test_doy_fraction_number(doy=d_args['doy'], expected=d_args['doy_frac']):
    assert float(utils.get_info(calcs.doy_fraction(ee.Number(doy)))) == pytest.approx(expected)

def test_doy_fraction_keyword(doy=d_args['doy'], expected=d_args['doy_frac']):
    output = utils.get_info(calcs.doy_fraction(doy=ee.Number(doy)))
    assert float(output) == pytest.approx(expected)

def test_doy_fraction_image(doy=d_args['doy'], expected=d_args['doy_frac']):
    output = utils.constant_image_value(calcs.doy_fraction(ee.Image.constant(doy)))
    assert float(output['constant']) == pytest.approx(expected)


def test_declination_default(doy=d_args['doy'], expected=d_args['delta_asce']):
    assert float(utils.get_info(calcs.declination(ee.Number(doy)))) == pytest.approx(expected)

def test_declination_asce(doy=d_args['doy'], expected=d_args['delta_asce']):
    output = utils.get_info(calcs.declination(ee.Number(doy), method='asce'))
    assert float(output) == pytest.approx(expected)

def test_declination_refet(doy=d_args['doy'], expected=d_args['delta']):
    output = utils.get_info(calcs.declination(ee.Number(doy), method='refet'))
    assert float(output) == pytest.approx(expected)

def test_declination_keyword(doy=d_args['doy'], expected=d_args['delta_asce']):
    assert float(utils.get_info(calcs.declination(doy=ee.Number(doy)))) == pytest.approx(expected)

def test_declination_image(doy=d_args['doy'], expected=d_args['delta_asce']):
    output = utils.constant_image_value(calcs.declination(ee.Image.constant(doy)))
    assert float(output['constant']) == pytest.approx(expected)


def test_dr_number(doy=d_args['doy'], expected=d_args['dr']):
    assert float(utils.get_info(calcs.dr(ee.Number(doy)))) == pytest.approx(expected)

def test_dr_keyword(doy=d_args['doy'], expected=d_args['dr']):
    assert float(utils.get_info(calcs.dr(doy=ee.Number(doy)))) == pytest.approx(expected)

def test_dr_image(doy=d_args['doy'], expected=d_args['dr']):
    output = utils.constant_image_value(calcs.dr(ee.Image.constant(doy)))
    assert float(output['constant']) == pytest.approx(expected)


def test_seasonal_correction_number(doy=d_args['doy'], expected=d_args['sc']):
    output = utils.get_info(calcs.seasonal_correction(ee.Number(doy)))
    assert float(output) == pytest.approx(expected)

def test_seasonal_correction_keyword(doy=d_args['doy'], expected=d_args['sc']):
    output = utils.get_info(calcs.seasonal_correction(doy=ee.Number(doy)))
    assert float(output) == pytest.approx(expected)

def test_seasonal_correction_image(doy=d_args['doy'], expected=d_args['sc']):
    output = utils.constant_image_value(calcs.seasonal_correction(ee.Image.constant(doy)))
    assert float(output['constant']) == pytest.approx(expected)


def test_solar_time_rad_number(
        lon=s_args['lon'], time_mid=h_args['time'], sc=d_args['sc'], expected=h_args['solar_time']
):
    output = utils.get_info(calcs.solar_time_rad(
        ee.Number(lon), ee.Number(time_mid), ee.Number(sc)
    ))
    assert float(output) == pytest.approx(expected)


def test_solar_time_rad_image(
        lon=s_args['lon'], time_mid=h_args['time'], sc=d_args['sc'], expected=h_args['solar_time']
):
    output = utils.constant_image_value(calcs.solar_time_rad(
        ee.Image.constant(lon), ee.Number(time_mid), ee.Number(sc)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# # The code will fail if the lon is a number but the time_mid is an image
# def test_solar_time_rad_image_fail(lon=s_args['lon'], time_mid=h_args['time'],
#                                    sc=d_args['sc'], expected=h_args['solar_time']):
#     output = utils.constant_image_value(calcs.solar_time_rad(
#         ee.Image.number(lon), ee.Image.constant(time_mid), ee.Number(sc)
#     ))
#     assert float(output['constant']) == pytest.approx(expected)


def test_sha_number(solar_time=h_args['solar_time'], expected=h_args['omega']):
    output = utils.get_info(calcs.sha(ee.Number(solar_time)))
    assert float(output) == pytest.approx(expected)

def test_sha_keyword(solar_time=h_args['solar_time'], expected=h_args['omega']):
    output = utils.get_info(calcs.sha(solar_time=ee.Number(solar_time)))
    assert float(output) == pytest.approx(expected)

def test_sha_image(solar_time=h_args['solar_time'], expected=h_args['omega']):
    output = utils.constant_image_value(calcs.sha(ee.Image.constant(solar_time)))
    assert float(output['constant']) == pytest.approx(expected)


@pytest.mark.parametrize(
    'x, x_min, x_max, expected',
    [
        [1.2, 1.2, 1.5, 1.2],
        [1.1, 1.2, 1.5, 1.4],
        [1.6, 1.2, 1.5, 1.3],
        [2.0, 1.2, 1.5, 1.4],
    ]
)
def test_wrap(x, x_min, x_max, expected):
    assert float(utils.get_info(calcs.wrap(ee.Number(x), x_min, x_max))) == pytest.approx(expected)


def test_sha_sunset_number(lat=s_args['lat'], delta=d_args['delta'], expected=d_args['omega_s']):
    output = utils.get_info(calcs.sha_sunset(lat=ee.Number(lat), delta=ee.Number(delta)))
    assert float(output) == pytest.approx(expected)

def test_sha_sunset_position(lat=s_args['lat'], delta=d_args['delta'], expected=d_args['omega_s']):
    output = utils.get_info(calcs.sha_sunset(ee.Number(lat), ee.Number(delta)))
    assert float(output) == pytest.approx(expected)

def test_sha_sunset_image(lat=s_args['lat'], delta=d_args['delta'], expected=d_args['omega_s']):
    output = utils.constant_image_value(calcs.sha_sunset(
        lat=ee.Image.constant(lat), delta=ee.Number(delta)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Ra Daily
def test_ra_daily_default(lat=s_args['lat'], doy=d_args['doy'], expected=d_args['ra_asce']):
    output = utils.get_info(calcs.ra_daily(lat=ee.Number(lat), doy=ee.Number(doy)))
    assert float(output) == pytest.approx(expected)

def test_ra_daily_asce(lat=s_args['lat'], doy=d_args['doy'], expected=d_args['ra_asce']):
    output = utils.get_info(calcs.ra_daily(
        lat=ee.Number(lat), doy=ee.Number(doy), method='asce'
    ))
    assert float(output) == pytest.approx(expected)

def test_ra_daily_refet(lat=s_args['lat'], doy=d_args['doy'], expected=d_args['ra']):
    output = utils.get_info(calcs.ra_daily(
        lat=ee.Number(lat), doy=ee.Number(doy), method='refet'
    ))
    assert float(output) == pytest.approx(expected)

def test_ra_daily_position(lat=s_args['lat'], doy=d_args['doy'], expected=d_args['ra_asce']):
    output = utils.get_info(calcs.ra_daily(ee.Number(lat), ee.Number(doy)))
    assert float(output) == pytest.approx(expected)

def test_ra_daily_image(lat=s_args['lat'], doy=d_args['doy'], expected=d_args['ra_asce']):
    output = utils.constant_image_value(calcs.ra_daily(
        lat=ee.Image.constant(lat), doy=ee.Number(doy)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Ra Hourly
def test_ra_hourly_default(lat=s_args['lat'], lon=s_args['lon'],
                           doy=h_args['doy'], time_mid=h_args['time_mid'],
                           expected=h_args['ra_asce']):
    output = utils.get_info(calcs.ra_hourly(
        lat=ee.Number(lat), lon=ee.Number(lon), doy=ee.Number(doy), time_mid=ee.Number(time_mid)
    ))
    assert float(output) == pytest.approx(expected)

def test_ra_hourly_asce(lat=s_args['lat'], lon=s_args['lon'],
                        doy=h_args['doy'], time_mid=h_args['time_mid'],
                        expected=h_args['ra_asce']):
    output = utils.get_info(calcs.ra_hourly(
        lat=ee.Number(lat), lon=ee.Number(lon), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), method='asce',
    ))
    assert float(output) == pytest.approx(expected)

def test_ra_hourly_refet(lat=s_args['lat'], lon=s_args['lon'],
                         doy=h_args['doy'], time_mid=h_args['time_mid'],
                         expected=h_args['ra']):
    output = utils.get_info(calcs.ra_hourly(
        lat=ee.Number(lat), lon=ee.Number(lon), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), method='refet',
    ))
    assert float(output) == pytest.approx(expected)

def test_ra_hourly_position(lat=s_args['lat'], lon=s_args['lon'],
                            doy=h_args['doy'], time_mid=h_args['time_mid'],
                            expected=h_args['ra_asce']):
    output = utils.get_info(calcs.ra_hourly(
        ee.Number(lat), ee.Number(lon), ee.Number(doy), ee.Number(time_mid)
    ))
    assert float(output) == pytest.approx(expected)

# def test_ra_hourly_image(lat=s_args['lat'], lon=s_args['lon'],
#                          doy=h_args['doy'], time_mid=h_args['time_mid'],
#                          expected=h_args['ra_asce']):
#     output = utils.constant_image_value(calcs.ra_hourly(
#         lat=ee.Number(lat), lon=ee.Number(lon), doy=ee.Number(doy),
#         time_mid=ee.Number(time_mid)
#     ))
#     assert float(output['constant']) == pytest.approx(expected)


# Rso Full Daily
def test_rso_daily_number(
        ea=d_args['ea'], pair=s_args['pair'], ra=d_args['ra'], doy=d_args['doy'],
        lat=s_args['lat'], expected=d_args['rso']
):
    output = utils.get_info(calcs.rso_daily(
        ea=ee.Number(ea), pair=ee.Number(pair), ra=ee.Number(ra), doy=ee.Number(doy),
        lat=ee.Number(lat),
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_daily_position(
        ea=d_args['ea'], pair=s_args['pair'], ra=d_args['ra'], doy=d_args['doy'],
        lat=s_args['lat'], expected=d_args['rso']
):
    output = utils.get_info(calcs.rso_daily(
        ee.Number(ea), ee.Number(pair), ee.Number(ra), ee.Number(doy), ee.Number(lat)
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_daily_image(
        ea=d_args['ea'], pair=s_args['pair'], ra=d_args['ra'], doy=d_args['doy'],
        lat=s_args['lat'], expected=d_args['rso']
):
    output = utils.constant_image_value(
        calcs.rso_daily(
            ea=ee.Image.constant(ea), pair=ee.Number(pair), ra=ee.Number(ra),
            doy=ee.Number(doy), lat=ee.Number(lat),
        )
    )
    assert float(output['constant']) == pytest.approx(expected)

def test_rso_daily_image_doy_number(
        ea=d_args['ea'], ra=d_args['ra'], pair=s_args['pair'], doy=d_args['doy'],
        lat=s_args['lat'], expected=d_args['rso']
):
    # Check that the function works correctly when DOY is a number but other inputs are images
    # The current formulation requires that pair be an image if lat is
    output = utils.constant_image_value(calcs.rso_daily(
        ea=ee.Image.constant(ea), pair=ee.Image.constant(pair), ra=ee.Number(ra),
        doy=ee.Number(doy), lat=ee.Image.constant(lat),
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Rso Full Hourly
def test_rso_hourly_default(ea=h_args['ea'], ra=h_args['ra'],
                            pair=s_args['pair'], doy=h_args['doy'],
                            time_mid=h_args['time_mid'], lat=s_args['lat'],
                            lon=s_args['lon'], expected=h_args['rso_asce']):
    output = utils.get_info(calcs.rso_hourly(
        ea=ee.Number(ea), ra=ee.Number(ra), pair=ee.Number(pair),
        doy=ee.Number(doy), time_mid=ee.Number(time_mid), lat=ee.Number(lat),
        lon=ee.Number(lon),
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_hourly_asce(ea=h_args['ea'], ra=h_args['ra'], pair=s_args['pair'],
                         doy=h_args['doy'], time_mid=h_args['time_mid'],
                         lat=s_args['lat'], lon=s_args['lon'],
                         expected=h_args['rso_asce']):
    output = utils.get_info(calcs.rso_hourly(
        ea=ee.Number(ea), ra=ee.Number(ra), pair=ee.Number(pair),
        doy=ee.Number(doy), time_mid=ee.Number(time_mid), lat=ee.Number(lat),
        lon=ee.Number(lon), method='asce',
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_hourly_refet(ea=h_args['ea'], ra=h_args['ra'],
                          pair=s_args['pair'], doy=h_args['doy'],
                          time_mid=h_args['time_mid'], lat=s_args['lat'],
                          lon=s_args['lon'], expected=h_args['rso']):
    output = utils.get_info(calcs.rso_hourly(
        ea=ee.Number(ea), ra=ee.Number(ra), pair=ee.Number(pair),
        doy=ee.Number(doy), time_mid=ee.Number(time_mid),
        lat=ee.Number(lat), lon=ee.Number(lon), method='refet',
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_hourly_position(ea=h_args['ea'], ra=h_args['ra'],
                             pair=s_args['pair'], doy=h_args['doy'],
                             time_mid=h_args['time_mid'], lat=s_args['lat'],
                             lon=s_args['lon'], expected=h_args['rso_asce']):
    output = utils.get_info(calcs.rso_hourly(
        ee.Number(ea), ee.Number(ra), ee.Number(pair), ee.Number(doy),
        ee.Number(time_mid), ee.Number(lat), ee.Number(lon)
    ))
    assert float(output) == pytest.approx(expected)

def test_rso_hourly_image(ea=h_args['ea'], ra=h_args['ra'],
                          pair=s_args['pair'],
                          doy=h_args['doy'], time_mid=h_args['time_mid'],
                          lat=s_args['lat'], lon=s_args['lon'],
                          expected=h_args['rso_asce']):
    output = utils.constant_image_value(calcs.rso_hourly(
        ea=ee.Image.constant(ea), ra=ee.Image.constant(ra), pair=ee.Number(pair),
        doy=ee.Number(doy), time_mid=ee.Number(time_mid), lat=ee.Number(lat),
        lon=ee.Number(lon),
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Rso Simple
@pytest.mark.parametrize(
    'ra, elev, expected',
    [
        [d_args['ra'], s_args['elev'], d_args['rso_simple']],
        [h_args['ra'], s_args['elev'], h_args['rso_simple']]
    ]
)
def test_rso_simple_number(ra, elev, expected):
    output = utils.get_info(calcs.rso_simple(ra=ee.Number(ra), elev=ee.Number(elev)))
    assert float(output) == pytest.approx(expected)

def test_rso_simple_position(ra=d_args['ra'], elev=s_args['elev'], expected=d_args['rso_simple']):
    output = utils.get_info(calcs.rso_simple(ee.Number(ra), ee.Number(elev)))
    assert float(output) == pytest.approx(expected)

def test_rso_simple_image(ra=d_args['ra'], elev=s_args['elev'], expected=d_args['rso_simple']):
    output = utils.constant_image_value(calcs.rso_simple(
        ra=ee.Image.constant(ra), elev=ee.Number(elev)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Fcd daily
def test_fcd_daily_number(rs=d_args['rs'], rso=d_args['rso'], expected=d_args['fcd']):
    output = utils.get_info(calcs.fcd_daily(rs=ee.Number(rs), rso=ee.Number(rso)))
    assert float(output) == pytest.approx(expected)

def test_fcd_daily_position(rs=d_args['rs'], rso=d_args['rso'], expected=d_args['fcd']):
    output = utils.get_info(calcs.fcd_daily(ee.Number(rs), ee.Number(rso)))
    assert float(output) == pytest.approx(expected)

def test_fcd_daily_image(rs=d_args['rs'], rso=d_args['rso'], expected=d_args['fcd']):
    output = utils.constant_image_value(calcs.fcd_daily(
        rs=ee.Image.constant(rs), rso=ee.Number(rso)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Fcd hourly
def test_fcd_hourly_default(rs=h_args['rs'], rso=h_args['rso'],
                            doy=h_args['doy'], time_mid=h_args['time_mid'],
                            lat=s_args['lat'], lon=s_args['lon'],
                            expected=h_args['fcd_asce']):
    output = utils.get_info(calcs.fcd_hourly(
        rs=ee.Number(rs), rso=ee.Number(rso), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), lat=ee.Number(lat), lon=ee.Number(lon),
    ))
    assert float(output) == pytest.approx(expected)

def test_fcd_hourly_asce(rs=h_args['rs'], rso=h_args['rso'],
                         doy=h_args['doy'], time_mid=h_args['time_mid'],
                         lat=s_args['lat'], lon=s_args['lon'],
                         expected=h_args['fcd_asce']):
    output = utils.get_info(calcs.fcd_hourly(
        rs=ee.Number(rs), rso=ee.Number(rso), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), lat=ee.Number(lat), lon=ee.Number(lon),
        method='asce',
    ))
    assert float(output) == pytest.approx(expected)

def test_fcd_hourly_refet(rs=h_args['rs'], rso=h_args['rso'],
                          doy=h_args['doy'], time_mid=h_args['time_mid'],
                          lat=s_args['lat'], lon=s_args['lon'],
                          expected=h_args['fcd']):
    output = utils.get_info(calcs.fcd_hourly(
        rs=ee.Number(rs), rso=ee.Number(rso), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), lat=ee.Number(lat), lon=ee.Number(lon),
        method='refet',
    ))
    assert float(output) == pytest.approx(expected)

def test_fcd_hourly_night(rs=h_args['rs'], rso=h_args['rso'], doy=h_args['doy'],
                          time_mid=6.5, lat=s_args['lat'], lon=s_args['lon'],
                          expected=1):
    # For now, check that nighttime fcd values are set to 1
    output = utils.get_info(calcs.fcd_hourly(
        rs=ee.Number(rs), rso=ee.Number(rso), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), lat=ee.Number(lat), lon=ee.Number(lon),
        method='refet',
    ))
    assert float(output) == pytest.approx(expected)

def test_fcd_hourly_position(rs=h_args['rs'], rso=h_args['rso'],
                             doy=h_args['doy'], time_mid=h_args['time_mid'],
                             lat=s_args['lat'], lon=s_args['lon'],
                             expected=h_args['fcd_asce']):
    output = utils.get_info(calcs.fcd_hourly(
        ee.Number(rs), ee.Number(rso), ee.Number(doy), ee.Number(time_mid),
        ee.Number(lat), ee.Number(lon),
    ))
    assert float(output) == pytest.approx(expected)

def test_fcd_hourly_image(rs=h_args['rs'], rso=h_args['rso'],
                          doy=h_args['doy'], time_mid=h_args['time_mid'],
                          lat=s_args['lat'], lon=s_args['lon'],
                          expected=h_args['fcd_asce']):
    output = utils.constant_image_value(calcs.fcd_hourly(
        rs=ee.Image.constant(rs), rso=ee.Number(rso), doy=ee.Number(doy),
        time_mid=ee.Number(time_mid), lat=ee.Number(lat), lon=ee.Number(lon),
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Rnl Daily
def test_rnl_daily_number(tmin=d_args['tmin'], tmax=d_args['tmax'],
                          ea=d_args['ea'], fcd=d_args['fcd'], expected=d_args['rnl']):
    output = utils.get_info(calcs.rnl_daily(
        tmax=ee.Number(tmax), tmin=ee.Number(tmin), ea=ee.Number(ea),
        fcd=ee.Number(fcd),
    ))
    assert float(output) == pytest.approx(expected)

def test_rnl_daily_position(tmin=d_args['tmin'], tmax=d_args['tmax'],
                            ea=d_args['ea'], fcd=d_args['fcd'],
                            expected=d_args['rnl']):
    output = utils.get_info(calcs.rnl_daily(
        ee.Number(tmax), ee.Number(tmin), ee.Number(ea), ee.Number(fcd)
    ))
    assert float(output) == pytest.approx(expected)

def test_rnl_daily_image(tmin=d_args['tmin'], tmax=d_args['tmax'],
                         ea=d_args['ea'], fcd=d_args['fcd'], expected=d_args['rnl']):
    output = utils.constant_image_value(calcs.rnl_daily(
        tmax=ee.Image.constant(tmax), tmin=ee.Number(tmin),
        ea=ee.Number(ea), fcd=ee.Number(fcd),
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Rnl Hourly
def test_rnl_hourly_number(tmean=h_args['tmean'], ea=h_args['ea'],
                           fcd=h_args['fcd'], expected=h_args['rnl']):
    output = utils.get_info(calcs.rnl_hourly(
        tmean=ee.Number(tmean), ea=ee.Number(ea), fcd=ee.Number(fcd)
    ))
    assert float(output) == pytest.approx(expected)

def test_rnl_hourly_position(tmean=h_args['tmean'], ea=h_args['ea'],
                             fcd=h_args['fcd'], expected=h_args['rnl']):
    output = utils.get_info(calcs.rnl_hourly(
        ee.Number(tmean), ee.Number(ea), ee.Number(fcd)
    ))
    assert float(output) == pytest.approx(expected)

def test_rnl_hourly_image(tmean=h_args['tmean'], ea=h_args['ea'],
                          fcd=h_args['fcd'], expected=h_args['rnl']):
    output = utils.constant_image_value(calcs.rnl_hourly(
        tmean=ee.Image.constant(tmean), ea=ee.Number(ea), fcd=ee.Number(fcd)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Rn
def test_rn_number(rs=d_args['rs'], rnl=d_args['rnl'], expected=d_args['rn']):
    output = utils.get_info(calcs.rn(rs=ee.Number(rs), rnl=ee.Number(rnl)))
    assert float(output) == pytest.approx(expected)

def test_rn_keywods(rs=d_args['rs'], rnl=d_args['rnl'], expected=d_args['rn']):
    output = utils.get_info(calcs.rn(ee.Number(rs), ee.Number(rnl)))
    assert float(output) == pytest.approx(expected)

def test_rn_image(rs=d_args['rs'], rnl=d_args['rnl'], expected=d_args['rn']):
    output = utils.constant_image_value(calcs.rn(
        rs=ee.Number(rs), rnl=ee.Image.constant(rnl)
    ))
    assert float(output['constant']) == pytest.approx(expected)


# Wind
def test_wind_height_adjust_2m(uz=2.5, zw=2.0, expected=2.5):
    output = utils.get_info(calcs.wind_height_adjust(uz=ee.Number(uz), zw=ee.Number(zw)))
    assert float(output) == pytest.approx(expected, abs=0.001)

def test_wind_height_adjust_position(uz=d_args['uz'], zw=s_args['zw'], expected=d_args['u2']):
    output = utils.get_info(calcs.wind_height_adjust(ee.Number(uz), ee.Number(zw)))
    assert float(output) == pytest.approx(expected)

@pytest.mark.parametrize(
    'uz, zw, expected',
    [
        [d_args['uz'], s_args['zw'], d_args['u2']],
        [h_args['uz'], s_args['zw'], h_args['u2']]
    ]
)
def test_wind_height_adjust_number(uz, zw, expected):
    output = utils.get_info(calcs.wind_height_adjust(uz=ee.Number(uz), zw=ee.Number(zw)))
    assert float(output) == pytest.approx(expected)

@pytest.mark.parametrize(
    'uz, zw, expected',
    [
        [d_args['uz'], s_args['zw'], d_args['u2']],
        [h_args['uz'], s_args['zw'], h_args['u2']]
    ]
)
def test_wind_height_adjust_image(uz, zw, expected):
    output = utils.constant_image_value(calcs.wind_height_adjust(
        uz=ee.Image.constant(uz), zw=ee.Number(zw)
    ))
    assert float(output['constant']) == pytest.approx(expected, abs=0.001)



import datetime as dt
from future.standard_library import install_aliases
install_aliases()

from urllib.request import urlopen


# Input parameters
station = 'FALN'
year = 2015
hourly_csv = '{}_Agrimet_hourly_raw_{}.csv'.format(station, year)
daily_csv = '{}_Agrimet_daily_raw_{}.csv'.format(station, year)



# Retrieve the data from the Agrimet site
print('Retrieving raw hourly data')
hourly_url = (
    "https://www.usbr.gov/pn-bin/instant.pl?"
    "station={0}&year={1}&month=1&day=1&year={1}&month=12&day=31&"
    "pcode=OB&pcode=TP&pcode=WS&pcode=SI&"
    # "pcode=OB&pcode=TU&pcode=EA&pcode=TP&pcode=WS&pcode=SI&"
    "print_hourly=1").format(station, year)
hourly_txt = urlopen(hourly_url).read().decode('utf-8')

# Select the data (i.e. strip off header and footer)
hourly_list = [x.strip() for x in hourly_txt.split('\n')]
start_i = hourly_list.index('BEGIN DATA')
end_i = hourly_list.index('END DATA')
hourly_list = hourly_list[start_i+1:end_i]

# Remove station name from column names
hourly_list[0] = hourly_list[0].replace(station, '').replace('DATE', 'DATE,')

# Clean up formatting
# This will probably fail if there are missing values
hourly_list = [
    ','.join([x.strip() for x in line.split(',')])
    for line in hourly_list]

# Change date to separate year, month, day columns
# Change time to hour
hourly_list[0] = hourly_list[0].replace('DATE,TIME', 'YEAR,MONTH,DAY,HOUR')
for i, line in enumerate(hourly_list[1:]):
    line_dt = line.split(',')[0]
    hourly_list[i+1] = line.replace(
        line_dt,
        dt.datetime.strptime(line_dt, '%m/%d/%Y %H:00').strftime('%Y,%m,%d,%H'))

# Write the raw data to a CSV file
print('Writing raw hourly data')
with open(hourly_csv, 'w') as f:
    for line in hourly_list:
        f.write(line + '\n')



# Retrieve the data from the Agrimet site
print('Retrieving raw daily data')
daily_url = (
    "https://www.usbr.gov/pn-bin/daily.pl?"
    "station={0}&year={1}&month=1&day=1&year={1}&month=12&day=31&"
    "pcode=MN&pcode=MX&pcode=SR&pcode=YM&pcode=UA&"
    "pcode=ETRS&pcode=ETOS").format(station, year)
daily_txt = urlopen(daily_url).read().decode('utf-8')

# Select the data (i.e. strip off header and footer)
daily_list = [x.strip() for x in daily_txt.split('\n')]
start_i = daily_list.index('BEGIN DATA')
end_i = daily_list.index('END DATA')
daily_list = daily_list[start_i+1:end_i]

# Remove station name from column names
daily_list[0] = daily_list[0].replace(station, '')

# Clean up formatting
# This will probably fail if there are missing values
daily_list = [
    ','.join([x.strip() for x in line.split(',')])
    for line in daily_list]

# Change date to separate year, month, day columns
# Change time to hour
daily_list[0] = daily_list[0].replace('DATE', 'YEAR,MONTH,DAY')
for i, line in enumerate(daily_list[1:]):
    line_dt = line.split(',')[0]
    daily_list[i+1] = line.replace(
        line_dt,
        dt.datetime.strptime(line_dt, '%m/%d/%Y').strftime('%Y,%m,%d'))

# Write the raw data to a CSV file
print('Writing raw daily data')
with open(daily_csv, 'w') as f:
    for line in daily_list:
        f.write(line + '\n')



import math


def deg2rad(deg):
    return deg * math.pi / 180.0


def rad2deg(rad):
    return rad * 180.0 / math.pi


def c2f(c):
    return c * (9.0 / 5) + 32


def f2c(f):
    return (f - 32) * (5.0 / 9)



from .daily import Daily
from .hourly import Hourly

# from importlib import metadata
# __version__ = metadata.version(__package__.replace('.', '-') or __name__.replace('.', '-'))
# # __version__ = metadata.version(__package__ or __name__)
# # __version__ = metadata.version('openet-refet-gee')




import ee
import pytest

from openet.refetgee import Daily
import openet.refetgee.units as units
import utils

# Eventually move to conftest.py or a separate file
# Fallon AgriMet site parameters
s_args = {
    'elev': 1208.5,
    'lat': 39.4575,
    'lon': -118.77388,
    'pair': 87.81876435813037,
    'pair_asce': 87.80710537212929,
    'zw': 3.0,
}

# Daily test parameters for 2015-07-01
# Ea was computed from q for both asce and refet methods
d_args = {
    'doy': 182,
    'ea': 1.2206674169951346,
    'eto_asce': 7.9422320475712835,
    'eto_refet': 7.9422320475712835,
    'etr_asce': 10.626087665395694,
    'etr_refet': 10.571314344056955,
    'etw_refet': 6.242411580074248,
    'eto_fs1': 4.373303512213786,
    'eto_fs2': 3.5689285353574958,
    'etr_rso_simple': 10.628137858930051,
    'q': 0.008691370735727117,          # Computed from Ea from Tdew
    'q_asce': 0.008692530868140688,     # Computed from Ea from Tdew
    'pet_hargreaves': 8.247962376780558,
    'rs': 674.07 * 0.041868,            # Conversion from Langleys to MJ m-2
    'rso': 31.565939444861765,
    'tdew': units.f2c(49.84),
    'tmin': units.f2c(66.65),
    'tmax': units.f2c(102.80),
    # 'tmean': f2c(84.725),
    'uz': 4.80 * 0.44704,               # Conversion from mph to m s-1
    'u2': 1.976111757722194,
}

# Compute wind x and y components
d_args['uz_x'] = d_args['uz'] / (2 ** 0.5)
d_args['uz_y'] = d_args['uz'] / (2 ** 0.5)
# d_args['uz_x'] = d_args['uz']
# d_args['uz_y'] = 0.0


# # Test full daily functions with positional inputs
# def test_refet_daily_input_positions():
#     refet = Daily(
#         ee.Image.constant(d_args['tmax']), ee.Image.constant(d_args['tmin']),
#         ee.Image.constant(d_args['ea']), ee.Image.constant(d_args['rs']),
#         ee.Image.constant(d_args['uz']), ee.Number(s_args['zw']),
#         ee.Number(s_args['elev']), ee.Number(s_args['lat']),
#         ee.Number(d_args['doy']), method='refet',
#     )
#     output = utils.constant_image_value(refet.etr)
#     assert float(output['etr']) == pytest.approx(d_args['etr'])


# Test full daily calculations with keyword inputs
# Test surface, rso_type, and rso inputs
def test_refet_daily_etr():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_refet'])


def test_refet_daily_eto():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.eto)

    assert float(output['eto']) == pytest.approx(d_args['eto_refet'])


def test_refet_daily_etw():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.etw)

    assert float(output['etw']) == pytest.approx(d_args['etw_refet'])


def test_refet_daily_eto_fs1():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.eto_fs1)

    assert float(output['eto_fs1']) == pytest.approx(d_args['eto_fs1'])


def test_refet_daily_eto_fs2():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.eto_fs2)

    assert float(output['eto_fs2']) == pytest.approx(d_args['eto_fs2'])


def test_refet_daily_pet_hargreaves():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(0), rs=ee.Image.constant(0),
        uz=ee.Image.constant(0), zw=ee.Number(s_args['zw']),
        elev=ee.Number(0), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='asce',
    )

    output = utils.constant_image_value(refet.pet_hargreaves)

    assert float(output['pet_hargreaves']) == pytest.approx(d_args['pet_hargreaves'])


def test_refet_daily_rso_type_simple():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet', rso_type='simple',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_rso_simple'])


def test_refet_daily_rso_type_array():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
        rso_type='array', rso=ee.Number(d_args['rso']),
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_refet'])


def test_refet_daily_rso_type_exception():
    with pytest.raises(ValueError):
        Daily(
            tmax=ee.Image.constant(d_args['tmax']),
            tmin=ee.Image.constant(d_args['tmin']),
            ea=ee.Image.constant(d_args['ea']),
            rs=ee.Image.constant(d_args['rs']),
            uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
            elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
            doy=ee.Number(d_args['doy']), rso_type='nonsense', method='refet',
        )


def test_refet_daily_etr_asce():
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


@pytest.mark.parametrize(
    'surface, expected',
    [
        ['etr', d_args['etr_refet']],
        ['alfalfa', d_args['etr_refet']],
        ['tall', d_args['etr_refet']],
        ['eto', d_args['eto_refet']],
        ['grass', d_args['eto_refet']],
        ['short', d_args['eto_refet']],
    ]
)
def test_refet_daily_etsz(surface, expected):
    refet = Daily(
        tmax=ee.Image.constant(d_args['tmax']), tmin=ee.Image.constant(d_args['tmin']),
        ea=ee.Image.constant(d_args['ea']), rs=ee.Image.constant(d_args['rs']),
        uz=ee.Image.constant(d_args['uz']), zw=ee.Number(s_args['zw']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        doy=ee.Number(d_args['doy']), method='refet',
    )

    output = utils.constant_image_value(refet.etsz(surface).rename(['etsz']))

    assert float(output['etsz']) == pytest.approx(expected)


def test_refet_daily_gridmet_etr():
    """Generate a mock GRIDMET image from the test values"""
    input_img = (
        ee.Image.constant([
            d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
            d_args['q_asce'], d_args['rs'] / 0.0864, d_args['uz']
        ])
        .rename(['tmmx', 'tmmn', 'sph', 'srad', 'vs'])
        .set('system:time_start', ee.Date('2015-07-01').millis())
    )

    refet = Daily.gridmet(
        ee.Image(input_img), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_maca_etr():
    """Generate a mock MACA image from the test values"""
    input_img = (
        ee.Image.constant([
            d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
            d_args['q_asce'], d_args['rs'] / 0.0864,
            d_args['uz_x'], d_args['uz_y']])
        .rename(['tasmax', 'tasmin', 'huss', 'rsds', 'uas', 'vas'])
        .set('system:time_start', ee.Date('2015-07-01').millis())
    )

    refet = Daily.maca(
        ee.Image(input_img), elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_nldas_etr():
    """Generate a mock NLDAS image from the test values

    Convert the test Rs from MJ m-2 d-1 to W m-2, then allocate half to each image

    """
    band_names = [
        'temperature', 'specific_humidity', 'shortwave_radiation', 'wind_u', 'wind_v'
    ]

    input_coll = ee.ImageCollection.fromImages([
        ee.Image.constant([d_args['tmin'], d_args['q_asce'], 0.0, d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T00:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'], d_args['q_asce'], d_args['rs'] / 0.0036,
                           d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T12:00:00', 'UTC').millis()})
    ])

    refet = Daily.nldas(
        input_coll, elev=ee.Number(s_args['elev']),
        lat=ee.Number(s_args['lat']), zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_nldas_eto():
    """Compare values to previous calculations"""
    test_point = ee.Geometry.Point(-120.113, 36.336)

    input_coll = (
        ee.ImageCollection('NASA/NLDAS/FORA0125_H002')
        .filterDate('2017-07-01T06:00:00', '2017-07-02T06:00:00')
    )
    output = utils.get_info(
        Daily.nldas(input_coll, method='asce').eto
        .reduceRegion(ee.Reducer.first(), geometry=test_point, scale=1)
    )
    expected = utils.get_info(
        ee.Image('projects/eddi-noaa/nldas/daily/20170701')
        .select(['eto_asce'])
        .reduceRegion(ee.Reducer.first(), geometry=test_point, scale=1)
    )

    assert output['eto'] == pytest.approx(expected['eto_asce'], rel=0.001)


def test_refet_daily_cfsv2_etr():
    """Generate a mock CFSv2 image from the test values

    Convert the test Rs from MJ m-2 d-1 to W m-2

    """
    band_names = [
        'Maximum_temperature_height_above_ground_6_Hour_Interval',
        'Minimum_temperature_height_above_ground_6_Hour_Interval',
        'Specific_humidity_height_above_ground',
        'Downward_Short-Wave_Radiation_Flux_surface_6_Hour_Average',
        'u-component_of_wind_height_above_ground',
        'v-component_of_wind_height_above_ground',
    ]

    input_coll = ee.ImageCollection.fromImages([
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
                           d_args['q_asce'], d_args['rs'] / 0.0864,
                           d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T00:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
                           d_args['q_asce'], d_args['rs'] / 0.0864,
                           d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T06:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
                           d_args['q_asce'], d_args['rs'] / 0.0864,
                           d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T12:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tmin'] + 273.15,
                           d_args['q_asce'], d_args['rs'] / 0.0864,
                           d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T18:00:00', 'UTC').millis()}),
    ])

    refet = Daily.cfsv2(
        input_coll, elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_rtma_etr():
    """Generate a mock RTMA image from the test values"""
    band_names = ['TMP', 'SPFH', 'WIND']

    input_coll = ee.ImageCollection.fromImages([
        ee.Image.constant([d_args['tmin'], d_args['q_asce'], d_args['uz']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T00:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'], d_args['q_asce'], d_args['uz']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T12:00:00', 'UTC').millis()})
    ])

    refet = Daily.rtma(
        input_coll, rs=ee.Image.constant(d_args['rs']),
        elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_era5_etr():
    """Generate a mock ERA5 image from the test values"""
    band_names = [
        'temperature_2m', 'dewpoint_temperature_2m',
        'surface_solar_radiation_downwards',
        'u_component_of_wind_10m', 'v_component_of_wind_10m',
    ]

    # Each image needs half of the total Rs since hourly images are summed
    input_coll = ee.ImageCollection.fromImages([
        ee.Image.constant([d_args['tmin'] + 273.15, d_args['tdew'] + 273.15,
                           0.0, d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T00:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tdew'] + 273.15,
                           d_args['rs'] * 1000000, d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T12:00:00', 'UTC').millis()})
    ])

    refet = Daily.era5(
        input_coll, elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_era5_land_etr():
    """Generate a mock ERA5-Land image from the test values"""
    band_names = [
        'temperature_2m', 'dewpoint_temperature_2m',
        'surface_solar_radiation_downwards_hourly',
        'u_component_of_wind_10m', 'v_component_of_wind_10m',
    ]

    # Each image needs half of the total Rs since hourly images are summed
    input_coll = ee.ImageCollection.fromImages([
        ee.Image.constant([d_args['tmin'] + 273.15, d_args['tdew'] + 273.15,
                           0.0, d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T00:00:00', 'UTC').millis()}),
        ee.Image.constant([d_args['tmax'] + 273.15, d_args['tdew'] + 273.15,
                           d_args['rs'] * 1000000, d_args['uz_x'], d_args['uz_y']])
            .double().rename(band_names)
            .set({'system:time_start': ee.Date('2015-07-01T12:00:00', 'UTC').millis()})
    ])

    refet = Daily.era5_land(
        input_coll, elev=ee.Number(s_args['elev']), lat=ee.Number(s_args['lat']),
        zw=ee.Number(s_args['zw']), method='asce',
    )

    output = utils.constant_image_value(refet.etr)

    assert float(output['etr']) == pytest.approx(d_args['etr_asce'])


def test_refet_daily_era5_land_fill_edge_cells():
    """Check that the fill_edge_cells flag works for an edge cell along the coast of England"""
    input_coll = ee.ImageCollection('ECMWF/ERA5_LAND/HOURLY').filterDate('2015-07-01', '2015-07-02')
    output = utils.point_image_value(Daily.era5_land(input_coll, fill_edge_cells=False).etr, xy=[0.0, 50.7])
    assert output['etr'] is None
    output = utils.point_image_value(Daily.era5_land(input_coll, fill_edge_cells=True).etr, xy=[0.0, 50.7])
    assert output['etr'] is not None


# TODO: Add a test for using the default Rs when one is not provided
